# -*- coding: utf-8 -*-
"""
reV supply curve module
- Calculation of LCOT
- Supply Curve creation
"""
import json
import logging
import os
from itertools import chain
from copy import deepcopy
from warnings import warn

import numpy as np
import pandas as pd
from rex import Resource
from rex.utilities import SpawnProcessPool, parse_table

from reV.handlers.transmission import TransmissionCosts as TC
from reV.handlers.transmission import TransmissionFeatures as TF
from reV.supply_curve.competitive_wind_farms import CompetitiveWindFarms
from reV.utilities import SupplyCurveField, log_versions
from reV.utilities.exceptions import SupplyCurveError, SupplyCurveInputError

logger = logging.getLogger(__name__)


# map is column name to relative order in which it should appear in output file
_REQUIRED_COMPUTE_AND_OUTPUT_COLS = {
    SupplyCurveField.TRANS_GID: 0,
    SupplyCurveField.TRANS_TYPE: 1,
    SupplyCurveField.N_PARALLEL_TRANS: 2,
    SupplyCurveField.DIST_SPUR_KM: 3,
    SupplyCurveField.TOTAL_TRANS_CAP_COST_PER_MW: 10,
    SupplyCurveField.LCOT: 11,
    SupplyCurveField.TOTAL_LCOE: 12,
}
_REQUIRED_OUTPUT_COLS = {SupplyCurveField.DIST_EXPORT_KM: 4,
                         SupplyCurveField.REINFORCEMENT_DIST_KM: 5,
                         SupplyCurveField.TIE_LINE_COST_PER_MW: 6,
                         SupplyCurveField.CONNECTION_COST_PER_MW: 7,
                         SupplyCurveField.EXPORT_COST_PER_MW: 8,
                         SupplyCurveField.REINFORCEMENT_COST_PER_MW: 9,
                         SupplyCurveField.POI_LAT: 13,
                         SupplyCurveField.POI_LON: 14,
                         SupplyCurveField.REINFORCEMENT_POI_LAT: 15,
                         SupplyCurveField.REINFORCEMENT_POI_LON: 16}
DEFAULT_COLUMNS = tuple(str(field)
                        for field in chain(_REQUIRED_COMPUTE_AND_OUTPUT_COLS,
                                           _REQUIRED_OUTPUT_COLS))
"""Default output columns from supply chain computation (not ordered)"""


class SupplyCurve:
    """SupplyCurve"""

    def __init__(self, sc_points, trans_table, sc_features=None,
                 # str() to fix docs
                 sc_capacity_col=str(SupplyCurveField.CAPACITY_AC_MW)):
        """ReV LCOT calculation and SupplyCurve sorting class.

        ``reV`` supply curve computes the transmission costs associated
        with each supply curve point output by ``reV`` supply curve
        aggregation. Transmission costs can either be computed
        competitively (where total capacity remaining on the
        transmission grid is tracked and updated after each new
        connection) or non-competitively (where the cheapest connections
        for each supply curve point are allowed regardless of the
        remaining transmission grid capacity). In both cases, the
        permutation of transmission costs between supply curve points
        and transmission grid features should be computed using the
        `reVX Least Cost Transmission Paths
        <https://github.com/NREL/reVX/tree/main/reVX/least_cost_xmission>`_
        utility.

        Parameters
        ----------
        sc_points : str | pandas.DataFrame
            Path to CSV or JSON or DataFrame containing supply curve
            point summary. Can also be a filepath to a ``reV`` bespoke
            HDF5 output file where the ``meta`` dataset has the same
            format as the supply curve aggregation output.

            .. Note:: If executing ``reV`` from the command line, this
              input can also be ``"PIPELINE"`` to parse the output of
              the previous pipeline step and use it as input to this
              call. However, note that duplicate executions of any
              preceding commands within the pipeline may invalidate this
              parsing, meaning the `sc_points` input will have to be
              specified manually.

        trans_table : str | pandas.DataFrame | list
            Path to CSV or JSON or DataFrame containing supply curve
            transmission mapping. This can also be a list of
            transmission tables with different line voltage (capacity)
            ratings. See the `reVX Least Cost Transmission Paths
            <https://github.com/NREL/reVX/tree/main/reVX/least_cost_xmission>`_
            utility to generate these input tables.
        sc_features : str | pandas.DataFrame, optional
            Path to CSV or JSON or DataFrame containing additional
            supply curve features (e.g. transmission multipliers,
            regions, etc.). These features will be merged to the
            `sc_points` input table on ALL columns that both have in
            common. If ``None``, no extra supply curve features are
            added. By default, ``None``.
        sc_capacity_col : str, optional
            Name of capacity column in `trans_sc_table`. The values in
            this column determine the size of transmission lines built.
            The transmission capital costs per MW and the reinforcement
            costs per MW will be returned in terms of these capacity
            values. Note that if this column != "capacity", then
            "capacity" must also be included in `trans_sc_table` since
            those values match the "mean_cf" data (which is used to
            calculate LCOT and Total LCOE). This input can be used to,
            e.g., size transmission lines based on solar AC capacity (
            ``sc_capacity_col="capacity_ac"``). By default,
            ``"capacity"``.

        Examples
        --------
        Standard outputs in addition to the values provided in
        `sc_points`, produced by
        :class:`reV.supply_curve.sc_aggregation.SupplyCurveAggregation`:

            - transmission_multiplier : int | float
                Transmission cost multiplier that scales the line cost
                but not the tie-in cost in the calculation of LCOT.
            - trans_gid : int
                Unique transmission feature identifier that each supply
                curve point was connected to.
            - trans_capacity : float
                Total capacity (not available capacity) of the
                transmission feature that each supply curve point was
                connected to. Default units are MW.
            - trans_type : str
                Tranmission feature type that each supply curve point
                was connected to (e.g. Transline, Substation).
            - trans_cap_cost_per_mw : float
                Capital cost of connecting each supply curve point to
                their respective transmission feature. This value
                includes line cost with transmission_multiplier and the
                tie-in cost. Default units are $/MW.
            - dist_km : float
                Distance in km from supply curve point to transmission
                connection.
            - lcot : float
                Levelized cost of connecting to transmission ($/MWh).
            - total_lcoe : float
                Total LCOE of each supply curve point (mean_lcoe + lcot)
                ($/MWh).
            - total_lcoe_friction : float
                Total LCOE of each supply curve point considering the
                LCOE friction scalar from the aggregation step
                (mean_lcoe_friction + lcot) ($/MWh).
        """
        log_versions(logger)
        logger.info("Supply curve points input: {}".format(sc_points))
        logger.info("Transmission table input: {}".format(trans_table))
        logger.info("Supply curve capacity column: {}".format(sc_capacity_col))

        self._sc_capacity_col = sc_capacity_col
        self._sc_points = self._parse_sc_points(
            sc_points, sc_features=sc_features
        )
        self._trans_table = self._map_tables(
            self._sc_points, trans_table, sc_capacity_col=sc_capacity_col
        )
        self._sc_gids, self._mask = self._parse_sc_gids(self._trans_table)

    def __repr__(self):
        msg = "{} with {} points".format(self.__class__.__name__, len(self))

        return msg

    def __len__(self):
        return len(self._sc_gids)

    def __getitem__(self, gid):
        if gid not in self._sc_gids:
            msg = "Invalid supply curve gid {}".format(gid)
            logger.error(msg)
            raise KeyError(msg)

        i = self._sc_gids.index(gid)

        return self._sc_points.iloc[i]

    @staticmethod
    def _parse_sc_points(sc_points, sc_features=None):
        """
        Import supply curve point summary and add any additional features

        Parameters
        ----------
        sc_points : str | pandas.DataFrame
            Path to .csv or .json or DataFrame containing supply curve point
            summary. Can also now be a filepath to a bespoke h5 where the
            "meta" dataset has the same format as the sc aggregation output.
        sc_features : str | pandas.DataFrame
            Path to .csv or .json or DataFrame containing additional supply
            curve features, e.g. transmission multipliers, regions

        Returns
        -------
        sc_points : pandas.DataFrame
            DataFrame of supply curve point summary with additional features
            added if supplied
        """
        if isinstance(sc_points, str) and sc_points.endswith(".h5"):
            with Resource(sc_points) as res:
                sc_points = res.meta
                sc_points.index.name = SupplyCurveField.SC_GID
                sc_points = sc_points.reset_index()
        else:
            sc_points = parse_table(sc_points)
            sc_points = sc_points.rename(
                columns=SupplyCurveField.map_from_legacy())

        logger.debug(
            "Supply curve points table imported with columns: {}".format(
                sc_points.columns.values.tolist()
            )
        )

        if sc_features is not None:
            sc_features = parse_table(sc_features)
            sc_features = sc_features.rename(
                columns=SupplyCurveField.map_from_legacy())
            merge_cols = [c for c in sc_features if c in sc_points]
            sc_points = sc_points.merge(sc_features, on=merge_cols, how="left")
            logger.debug(
                "Adding Supply Curve Features table with columns: {}".format(
                    sc_features.columns.values.tolist()
                )
            )

        if "transmission_multiplier" in sc_points:
            col = "transmission_multiplier"
            sc_points.loc[:, col] = sc_points.loc[:, col].fillna(1)

        logger.debug(
            "Final supply curve points table has columns: {}".format(
                sc_points.columns.values.tolist()
            )
        )

        return sc_points

    @staticmethod
    def _get_merge_cols(sc_columns, trans_columns):
        """
        Get columns with 'row' or 'col' in them to use for merging

        Parameters
        ----------
        sc_columns : list
            Columns to search
        trans_cols

        Returns
        -------
        merge_cols : dict
            Columns to merge on which maps the sc columns (keys) to the
            corresponding trans table columns (values)
        """
        sc_columns = [c for c in sc_columns if c.startswith("sc_")]
        trans_columns = [c for c in trans_columns if c.startswith("sc_")]
        merge_cols = {}
        for c_val in ["row", "col"]:
            trans_col = [c for c in trans_columns if c_val in c]
            sc_col = [c for c in sc_columns if c_val in c]
            if trans_col and sc_col:
                merge_cols[sc_col[0]] = trans_col[0]

        if len(merge_cols) != 2:
            msg = (
                "Did not find a unique set of sc row and column ids to "
                "merge on: {}".format(merge_cols)
            )
            logger.error(msg)
            raise RuntimeError(msg)

        return merge_cols

    @staticmethod
    def _parse_trans_table(trans_table):
        """
        Import transmission features table

        Parameters
        ----------
        trans_table : pd.DataFrame | str
            Table mapping supply curve points to transmission features
            (either str filepath to table file or pre-loaded dataframe).

        Returns
        -------
        trans_table : pd.DataFrame
            Loaded transmission feature table.
        """

        trans_table = parse_table(trans_table)

        # Update legacy transmission table columns to match new less ambiguous
        # column names:
        # trans_gid -> the transmission feature id, legacy name: trans_line_gid
        # trans_line_gids -> gids of transmission lines connected to the given
        # transmission feature (only used for Substations),
        # legacy name: trans_gids
        # also xformer_cost_p_mw -> xformer_cost_per_mw (not sure why there
        # would be a *_p_mw but here we are...)
        rename_map = {
            "trans_line_gid": SupplyCurveField.TRANS_GID,
            "trans_gids": "trans_line_gids",
            "xformer_cost_p_mw": "xformer_cost_per_mw",
        }
        trans_table = trans_table.rename(columns=rename_map)

        contains_dist_in_miles = "dist_mi" in trans_table
        missing_km_dist = SupplyCurveField.DIST_SPUR_KM not in trans_table
        if contains_dist_in_miles and missing_km_dist:
            trans_table = trans_table.rename(
                columns={"dist_mi": SupplyCurveField.DIST_SPUR_KM}
            )
            trans_table[SupplyCurveField.DIST_SPUR_KM] *= 1.60934

        drop_cols = [SupplyCurveField.SC_GID, 'cap_left',
                     SupplyCurveField.SC_POINT_GID]
        drop_cols = [c for c in drop_cols if c in trans_table]
        if drop_cols:
            trans_table = trans_table.drop(columns=drop_cols)

        return trans_table.rename(columns=SupplyCurveField.map_from_legacy())

    @staticmethod
    def _map_trans_capacity(trans_sc_table,
                            sc_capacity_col=SupplyCurveField.CAPACITY_AC_MW):
        """
        Map SC gids to transmission features based on capacity. For any SC
        gids with capacity > the maximum transmission feature capacity, map
        SC gids to the feature with the largest capacity

        Parameters
        ----------
        trans_sc_table : pandas.DataFrame
            Table mapping supply curve points to transmission features.
        sc_capacity_col : str, optional
            Name of capacity column in `trans_sc_table`. The values in
            this column determine the size of transmission lines built.
            The transmission capital costs per MW and the reinforcement
            costs per MW will be returned in terms of these capacity
            values. Note that if this column != "capacity", then
            "capacity" must also be included in `trans_sc_table` since
            those values match the "mean_cf" data (which is used to
            calculate LCOT and Total LCOE). By default, ``"capacity"``.

        Returns
        -------
        trans_sc_table : pandas.DataFrame
            Updated table mapping supply curve points to transmission features
            based on maximum capacity
        """

        nx = trans_sc_table[sc_capacity_col] / trans_sc_table["max_cap"]
        nx = np.ceil(nx).astype(int)
        trans_sc_table[SupplyCurveField.N_PARALLEL_TRANS] = nx

        if (nx > 1).any():
            mask = nx > 1
            tie_line_cost = (
                trans_sc_table.loc[mask, "tie_line_cost"] * nx[mask]
            )

            xformer_cost = (
                trans_sc_table.loc[mask, "xformer_cost_per_mw"]
                * trans_sc_table.loc[mask, "max_cap"]
                * nx[mask]
            )

            conn_cost = (
                xformer_cost
                + trans_sc_table.loc[mask, "sub_upgrade_cost"]
                + trans_sc_table.loc[mask, "new_sub_cost"]
            )

            trans_cap_cost = tie_line_cost + conn_cost

            trans_sc_table.loc[mask, "tie_line_cost"] = tie_line_cost
            trans_sc_table.loc[mask, "xformer_cost"] = xformer_cost
            trans_sc_table.loc[mask, "connection_cost"] = conn_cost
            trans_sc_table.loc[mask, "trans_cap_cost"] = trans_cap_cost

            msg = (
                "{} SC points have a capacity that exceeds the maximum "
                "transmission feature capacity and will be connected with "
                "multiple parallel transmission features.".format(
                    (nx > 1).sum()
                )
            )
            logger.info(msg)

        return trans_sc_table

    @staticmethod
    def _parse_trans_line_gids(trans_line_gids):
        """
        Parse json string of trans_line_gids if needed

        Parameters
        ----------
        trans_line_gids : str | list
            list of transmission line 'trans_gid's, if a json string, convert
            to list

        Returns
        -------
        trans_line_gids : list
            list of transmission line 'trans_gid's
        """
        if isinstance(trans_line_gids, str):
            trans_line_gids = json.loads(trans_line_gids)

        return trans_line_gids

    @classmethod
    def _check_sub_trans_lines(cls, features):
        """
        Check to make sure all trans-lines are available for all sub-stations

        Parameters
        ----------
        features : pandas.DataFrame
            Table of transmission feature to check substation to transmission
            line gid connections

        Returns
        -------
        line_gids : list
            List of missing transmission line 'trans_gid's for all substations
            in features table
        """
        features = features.rename(
            columns={
                "trans_line_gid": SupplyCurveField.TRANS_GID,
                "trans_gids": "trans_line_gids",
            }
        )
        mask = (features[SupplyCurveField.TRANS_TYPE].str.casefold()
                == "substation")

        if not any(mask):
            return []

        line_gids = features.loc[mask, "trans_line_gids"].apply(
            cls._parse_trans_line_gids
        )

        line_gids = np.unique(np.concatenate(line_gids.values))

        test = np.isin(line_gids, features[SupplyCurveField.TRANS_GID].values)

        return line_gids[~test].tolist()

    @classmethod
    def _check_substation_conns(cls, trans_table,
                                sc_cols=SupplyCurveField.SC_GID):
        """
        Run checks on substation transmission features to make sure that
        every sc point connecting to a substation can also connect to its
        respective transmission lines

        Parameters
        ----------
        trans_table : pd.DataFrame
            Table mapping supply curve points to transmission features
            (should already be merged with SC points).
        sc_cols : str | list, optional
            Column(s) in trans_table with unique supply curve id,
            by default SupplyCurveField.SC_GID
        """
        missing = {}
        for sc_point, sc_table in trans_table.groupby(sc_cols):
            tl_gids = cls._check_sub_trans_lines(sc_table)
            if tl_gids:
                missing[sc_point] = tl_gids

        if any(missing):
            msg = (
                "The following sc_gid (keys) were connected to substations "
                "but were not connected to the respective transmission line"
                " gids (values) which is required for full SC sort: {}".format(
                    missing
                )
            )
            logger.error(msg)
            raise SupplyCurveInputError(msg)

    @classmethod
    def _check_sc_trans_table(cls, sc_points, trans_table):
        """Run self checks on sc_points table and the merged trans_table

        Parameters
        ----------
        sc_points : pd.DataFrame
            Table of supply curve point summary
        trans_table : pd.DataFrame
            Table mapping supply curve points to transmission features
            (should already be merged with SC points).
        """
        sc_gids = set(sc_points[SupplyCurveField.SC_GID].unique())
        trans_sc_gids = set(trans_table[SupplyCurveField.SC_GID].unique())
        missing = sorted(list(sc_gids - trans_sc_gids))
        if any(missing):
            msg = (
                "There are {} Supply Curve points with missing "
                "transmission mappings. Supply curve points with no "
                "transmission features will not be connected! "
                "Missing sc_gid's: {}".format(len(missing), missing)
            )
            logger.warning(msg)
            warn(msg)

        if not any(trans_sc_gids) or not any(sc_gids):
            msg = (
                "Merging of sc points table and transmission features "
                "table failed with {} original sc gids and {} transmission "
                "sc gids after table merge.".format(
                    len(sc_gids), len(trans_sc_gids)
                )
            )
            logger.error(msg)
            raise SupplyCurveError(msg)

        logger.debug(
            "There are {} original SC gids and {} sc gids in the "
            "merged transmission table.".format(
                len(sc_gids), len(trans_sc_gids)
            )
        )
        logger.debug(
            "Transmission Table created with columns: {}".format(
                trans_table.columns.values.tolist()
            )
        )

    @classmethod
    def _merge_sc_trans_tables(cls, sc_points, trans_table,
                               sc_cols=(SupplyCurveField.SC_GID,
                                        SupplyCurveField.CAPACITY_AC_MW,
                                        SupplyCurveField.MEAN_CF_AC,
                                        SupplyCurveField.MEAN_LCOE),
                               sc_capacity_col=SupplyCurveField.CAPACITY_AC_MW
                               ):
        """
        Merge the supply curve table with the transmission features table.

        Parameters
        ----------
        sc_points : pd.DataFrame
            Table of supply curve point summary
        trans_table : pd.DataFrame | str
            Table mapping supply curve points to transmission features
            (either str filepath to table file, list of filepaths to tables by
             line voltage (capacity) or pre-loaded dataframe).
        sc_cols : tuple | list, optional
            List of column from sc_points to transfer into the trans table,
            If the `sc_capacity_col` is not included, it will get added.
            by default (SupplyCurveField.SC_GID, 'capacity', 'mean_cf',
            'mean_lcoe')
        sc_capacity_col : str, optional
            Name of capacity column in `trans_sc_table`. The values in
            this column determine the size of transmission lines built.
            The transmission capital costs per MW and the reinforcement
            costs per MW will be returned in terms of these capacity
            values. Note that if this column != "capacity", then
            "capacity" must also be included in `trans_sc_table` since
            those values match the "mean_cf" data (which is used to
            calculate LCOT and Total LCOE). By default, ``"capacity"``.

        Returns
        -------
        trans_sc_table : pd.DataFrame
            Updated table mapping supply curve points to transmission features.
            This is performed by an inner merging with trans_table
        """
        if sc_capacity_col not in sc_cols:
            sc_cols = tuple([sc_capacity_col] + list(sc_cols))

        if isinstance(trans_table, (list, tuple)):
            trans_sc_table = []
            for table in trans_table:
                trans_sc_table.append(
                    cls._merge_sc_trans_tables(
                        sc_points,
                        table,
                        sc_cols=sc_cols,
                        sc_capacity_col=sc_capacity_col,
                    )
                )

            trans_sc_table = pd.concat(trans_sc_table)
        else:
            trans_table = cls._parse_trans_table(trans_table)

            merge_cols = cls._get_merge_cols(
                sc_points.columns, trans_table.columns
            )
            logger.info(
                "Merging SC table and Trans Table with "
                "{} mapping: {}".format(
                    "sc_table_col: trans_table_col", merge_cols
                )
            )
            sc_points = sc_points.rename(columns=merge_cols)
            merge_cols = list(merge_cols.values())

            if isinstance(sc_cols, tuple):
                sc_cols = list(sc_cols)

            extra_cols = [SupplyCurveField.CAPACITY_DC_MW,
                          SupplyCurveField.MEAN_CF_DC,
                          SupplyCurveField.MEAN_LCOE_FRICTION,
                          "transmission_multiplier"]
            for col in extra_cols:
                if col in sc_points:
                    sc_cols.append(col)

            sc_cols += merge_cols
            sc_points = sc_points[sc_cols].copy()
            trans_sc_table = trans_table.merge(
                sc_points, on=merge_cols, how="inner"
            )

        return trans_sc_table

    @classmethod
    def _map_tables(cls, sc_points, trans_table,
                    sc_cols=(SupplyCurveField.SC_GID,
                             SupplyCurveField.CAPACITY_AC_MW,
                             SupplyCurveField.MEAN_CF_AC,
                             SupplyCurveField.MEAN_LCOE),
                    sc_capacity_col=SupplyCurveField.CAPACITY_AC_MW):
        """
        Map supply curve points to transmission features

        Parameters
        ----------
        sc_points : pd.DataFrame
            Table of supply curve point summary
        trans_table : pd.DataFrame | str
            Table mapping supply curve points to transmission features
            (either str filepath to table file, list of filepaths to tables by
             line voltage (capacity) or pre-loaded DataFrame).
        sc_cols : tuple | list, optional
            List of column from sc_points to transfer into the trans table,
            If the `sc_capacity_col` is not included, it will get added.
            by default (SupplyCurveField.SC_GID,
            SupplyCurveField.CAPACITY_AC_MW, SupplyCurveField.MEAN_CF_AC,
            SupplyCurveField.MEAN_LCOE)
        sc_capacity_col : str, optional
            Name of capacity column in `trans_sc_table`. The values in
            this column determine the size of transmission lines built.
            The transmission capital costs per MW and the reinforcement
            costs per MW will be returned in terms of these capacity
            values. Note that if this column != "capacity", then
            "capacity" must also be included in `trans_sc_table` since
            those values match the "mean_cf" data (which is used to
            calculate LCOT and Total LCOE). By default, ``"capacity"``.

        Returns
        -------
        trans_sc_table : pd.DataFrame
            Updated table mapping supply curve points to transmission features.
            This is performed by an inner merging with trans_table
        """
        scc = sc_capacity_col
        trans_sc_table = cls._merge_sc_trans_tables(
            sc_points, trans_table, sc_cols=sc_cols, sc_capacity_col=scc
        )

        if "max_cap" in trans_sc_table:
            trans_sc_table = cls._map_trans_capacity(
                trans_sc_table, sc_capacity_col=scc
            )

        sort_cols = [SupplyCurveField.SC_GID, SupplyCurveField.TRANS_GID]
        trans_sc_table = trans_sc_table.sort_values(sort_cols)
        trans_sc_table = trans_sc_table.reset_index(drop=True)

        cls._check_sc_trans_table(sc_points, trans_sc_table)

        return trans_sc_table

    @staticmethod
    def _create_handler(trans_table, trans_costs=None, avail_cap_frac=1):
        """
        Create TransmissionFeatures handler from supply curve transmission
        mapping table.  Update connection costs if given.

        Parameters
        ----------
        trans_table : str | pandas.DataFrame
            Path to .csv or .json or DataFrame containing supply curve
            transmission mapping
        trans_costs : str | dict
            Transmission feature costs to use with TransmissionFeatures
            handler: line_tie_in_cost, line_cost, station_tie_in_cost,
            center_tie_in_cost, sink_tie_in_cost
        avail_cap_frac: int, optional
            Fraction of transmissions features capacity 'ac_cap' to make
            available for connection to supply curve points, by default 1

        Returns
        -------
        trans_features : TransmissionFeatures
            TransmissionFeatures or TransmissionCosts instance initilized
            with specified transmission costs
        """
        if trans_costs is not None:
            kwargs = TF._parse_dictionary(trans_costs)
        else:
            kwargs = {}

        trans_features = TF(
            trans_table, avail_cap_frac=avail_cap_frac, **kwargs
        )

        return trans_features

    @staticmethod
    def _parse_sc_gids(trans_table, gid_key=SupplyCurveField.SC_GID):
        """Extract unique sc gids, make bool mask from tranmission table

        Parameters
        ----------
        trans_table : pd.DataFrame
            reV Supply Curve table joined with transmission features table.
        gid_key : str
            Column label in trans_table containing the supply curve points
            primary key.

        Returns
        -------
        sc_gids : list
            List of unique integer supply curve gids (non-nan)
        mask : np.ndarray
            Boolean array initialized as true. Length is equal to the maximum
            SC gid so that the SC gids can be used to index the mask directly.
        """
        sc_gids = list(np.sort(trans_table[gid_key].unique()))
        sc_gids = [int(gid) for gid in sc_gids]
        mask = np.ones(int(1 + max(sc_gids)), dtype=bool)

        return sc_gids, mask

    @staticmethod
    def _get_capacity(sc_gid, sc_table, connectable=True,
                      sc_capacity_col=SupplyCurveField.CAPACITY_AC_MW):
        """
        Get capacity of supply curve point

        Parameters
        ----------
        sc_gid : int
            Supply curve gid
        sc_table : pandas.DataFrame
            DataFrame of sc point to transmission features mapping for given
            sc_gid
        connectable : bool, optional
            Flag to ensure SC point can connect to transmission features,
            by default True
        sc_capacity_col : str, optional
            Name of capacity column in `trans_sc_table`. The values in
            this column determine the size of transmission lines built.
            The transmission capital costs per MW and the reinforcement
            costs per MW will be returned in terms of these capacity
            values. Note that if this column != "capacity", then
            "capacity" must also be included in `trans_sc_table` since
            those values match the "mean_cf" data (which is used to
            calculate LCOT and Total LCOE). By default, ``"capacity"``.

        Returns
        -------
        capacity : float
            Capacity of supply curve point
        """
        if connectable:
            capacity = sc_table[sc_capacity_col].unique()
            if len(capacity) == 1:
                capacity = capacity[0]
            else:
                msg = (
                    "Each supply curve point should only have "
                    "a single capacity, but {} has {}".format(sc_gid, capacity)
                )
                logger.error(msg)
                raise RuntimeError(msg)
        else:
            capacity = None

        return capacity

    @classmethod
    def _compute_trans_cap_cost(cls, trans_table, trans_costs=None,
                                avail_cap_frac=1, max_workers=None,
                                connectable=True, line_limited=False,
                                sc_capacity_col=(
                                    SupplyCurveField.CAPACITY_AC_MW)):
        """
        Compute levelized cost of transmission for all combinations of
        supply curve points and tranmission features in trans_table

        Parameters
        ----------
        trans_table : pd.DataFrame
            Table mapping supply curve points to transmission features
            MUST contain `sc_capacity_col` column.
        fcr : float
            Fixed charge rate needed to compute LCOT
        trans_costs : str | dict
            Transmission feature costs to use with TransmissionFeatures
            handler: line_tie_in_cost, line_cost, station_tie_in_cost,
            center_tie_in_cost, sink_tie_in_cost
        avail_cap_frac: int, optional
            Fraction of transmissions features capacity 'ac_cap' to make
            available for connection to supply curve points, by default 1
        max_workers : int | NoneType
            Number of workers to use to compute lcot, if > 1 run in parallel.
            None uses all available cpu's.
        connectable : bool, optional
            Flag to only compute tranmission capital cost if transmission
            feature has enough available capacity, by default True
        line_limited : bool
            Substation connection is limited by maximum capacity of the
            attached lines, legacy method
        sc_capacity_col : str, optional
            Name of capacity column in `trans_sc_table`. The values in
            this column determine the size of transmission lines built.
            The transmission capital costs per MW and the reinforcement
            costs per MW will be returned in terms of these capacity
            values. Note that if this column != "capacity", then
            "capacity" must also be included in `trans_sc_table` since
            those values match the "mean_cf" data (which is used to
            calculate LCOT and Total LCOE). By default, ``"capacity"``.

        Returns
        -------
        lcot : list
            Levelized cost of transmission for all supply curve -
            tranmission feature connections
        cost : list
            Capital cost of tramsmission for all supply curve - transmission
            feature connections
        """
        scc = sc_capacity_col
        if scc not in trans_table:
            raise SupplyCurveInputError(
                "Supply curve table must have "
                "supply curve point capacity column"
                "({}) to compute lcot".format(scc)
            )

        if trans_costs is not None:
            trans_costs = TF._parse_dictionary(trans_costs)
        else:
            trans_costs = {}

        if max_workers is None:
            max_workers = os.cpu_count()

        logger.info('Computing LCOT costs for all possible connections...')
        groups = trans_table.groupby(SupplyCurveField.SC_GID)
        if max_workers > 1:
            loggers = [__name__, "reV.handlers.transmission", "reV"]
            with SpawnProcessPool(
                max_workers=max_workers, loggers=loggers
            ) as exe:
                futures = []
                for sc_gid, sc_table in groups:
                    capacity = cls._get_capacity(
                        sc_gid,
                        sc_table,
                        connectable=connectable,
                        sc_capacity_col=scc,
                    )
                    futures.append(
                        exe.submit(
                            TC.feature_costs,
                            sc_table,
                            capacity=capacity,
                            avail_cap_frac=avail_cap_frac,
                            line_limited=line_limited,
                            **trans_costs,
                        )
                    )

                cost = [future.result() for future in futures]
        else:
            cost = []
            for sc_gid, sc_table in groups:
                capacity = cls._get_capacity(
                    sc_gid,
                    sc_table,
                    connectable=connectable,
                    sc_capacity_col=scc,
                )
                cost.append(
                    TC.feature_costs(
                        sc_table,
                        capacity=capacity,
                        avail_cap_frac=avail_cap_frac,
                        line_limited=line_limited,
                        **trans_costs,
                    )
                )

        cost = np.hstack(cost).astype("float32")
        logger.info("LCOT cost calculation is complete.")

        return cost

    def compute_total_lcoe(
        self,
        fcr,
        transmission_costs=None,
        avail_cap_frac=1,
        line_limited=False,
        connectable=True,
        max_workers=None,
        consider_friction=True,
    ):
        """
        Compute LCOT and total LCOE for all sc point to transmission feature
        connections

        Parameters
        ----------
        fcr : float
            Fixed charge rate, used to compute LCOT
        transmission_costs : str | dict, optional
            Transmission feature costs to use with TransmissionFeatures
            handler: line_tie_in_cost, line_cost, station_tie_in_cost,
            center_tie_in_cost, sink_tie_in_cost, by default None
        avail_cap_frac : int, optional
            Fraction of transmissions features capacity 'ac_cap' to make
            available for connection to supply curve points, by default 1
        line_limited : bool, optional
            Flag to have substation connection is limited by maximum capacity
            of the attached lines, legacy method, by default False
        connectable : bool, optional
            Flag to only compute tranmission capital cost if transmission
            feature has enough available capacity, by default True
        max_workers : int | NoneType, optional
            Number of workers to use to compute lcot, if > 1 run in parallel.
            None uses all available cpu's. by default None
        consider_friction : bool, optional
            Flag to consider friction layer on LCOE when "mean_lcoe_friction"
            is in the sc points input, by default True
        """
        tcc_per_mw_col = SupplyCurveField.TOTAL_TRANS_CAP_COST_PER_MW
        if tcc_per_mw_col in self._trans_table:
            cost = self._trans_table[tcc_per_mw_col].values.copy()
        elif "trans_cap_cost" not in self._trans_table:
            scc = self._sc_capacity_col
            cost = self._compute_trans_cap_cost(
                self._trans_table,
                trans_costs=transmission_costs,
                avail_cap_frac=avail_cap_frac,
                line_limited=line_limited,
                connectable=connectable,
                max_workers=max_workers,
                sc_capacity_col=scc,
            )
            self._trans_table[tcc_per_mw_col] = cost  # $/MW
        else:
            cost = self._trans_table["trans_cap_cost"].values.copy()  # $
            cost /= self._trans_table[SupplyCurveField.CAPACITY_AC_MW]  # $/MW
            self._trans_table[tcc_per_mw_col] = cost

        self._trans_table[tcc_per_mw_col] = (
            self._trans_table[tcc_per_mw_col].astype("float32")
        )
        cost = cost.astype("float32")
        cf_mean_arr = self._trans_table[SupplyCurveField.MEAN_CF_AC]
        cf_mean_arr = cf_mean_arr.values.astype("float32")
        resource_lcoe = self._trans_table[SupplyCurveField.MEAN_LCOE]
        resource_lcoe = resource_lcoe.values.astype("float32")

        if 'reinforcement_cost_floored_per_mw' in self._trans_table:
            logger.info("'reinforcement_cost_floored_per_mw' column found in "
                        "transmission table. Adding floored reinforcement "
                        "cost LCOE as sorting option.")
            fr_cost = (self._trans_table['reinforcement_cost_floored_per_mw']
                       .values.copy())

            lcot_fr = ((cost + fr_cost) * fcr) / (cf_mean_arr * 8760)
            lcoe_fr = lcot_fr + resource_lcoe
            self._trans_table['lcot_floored_reinforcement'] = lcot_fr
            self._trans_table['lcoe_floored_reinforcement'] = lcoe_fr

        if SupplyCurveField.REINFORCEMENT_COST_PER_MW in self._trans_table:
            logger.info("%s column found in transmission table. Adding "
                        "reinforcement costs to total LCOE.",
                        SupplyCurveField.REINFORCEMENT_COST_PER_MW)
            lcot_nr = (cost * fcr) / (cf_mean_arr * 8760)
            lcoe_nr = lcot_nr + resource_lcoe
            self._trans_table['lcot_no_reinforcement'] = lcot_nr
            self._trans_table['lcoe_no_reinforcement'] = lcoe_nr

            col_name = SupplyCurveField.REINFORCEMENT_COST_PER_MW
            r_cost = self._trans_table[col_name].astype("float32")
            r_cost = r_cost.values.copy()
            self._trans_table[tcc_per_mw_col] += r_cost
            cost += r_cost  # $/MW

        lcot = (cost * fcr) / (cf_mean_arr * 8760)
        self._trans_table[SupplyCurveField.LCOT] = lcot
        self._trans_table[SupplyCurveField.TOTAL_LCOE] = lcot + resource_lcoe

        if consider_friction:
            self._calculate_total_lcoe_friction()

    def _calculate_total_lcoe_friction(self):
        """Look for site mean LCOE with friction in the trans table and if
        found make a total LCOE column with friction."""

        if SupplyCurveField.MEAN_LCOE_FRICTION in self._trans_table:
            lcoe_friction = (
                self._trans_table[SupplyCurveField.LCOT]
                + self._trans_table[SupplyCurveField.MEAN_LCOE_FRICTION])
            self._trans_table[SupplyCurveField.TOTAL_LCOE_FRICTION] = (
                lcoe_friction
            )
            logger.info('Found mean LCOE with friction. Adding key '
                        '"total_lcoe_friction" to trans table.')

    def _exclude_noncompetitive_wind_farms(
        self, comp_wind_dirs, sc_gid, downwind=False
    ):
        """
        Exclude non-competitive wind farms for given sc_gid

        Parameters
        ----------
        comp_wind_dirs : CompetitiveWindFarms
            Pre-initilized CompetitiveWindFarms instance
        sc_gid : int
            Supply curve gid to exclude non-competitive wind farms around
        downwind : bool, optional
            Flag to remove downwind neighbors as well as upwind neighbors,
            by default False

        Returns
        -------
        comp_wind_dirs : CompetitiveWindFarms
            updated CompetitiveWindFarms instance
        """
        gid = comp_wind_dirs.check_sc_gid(sc_gid)
        if gid is not None:
            if comp_wind_dirs.mask[gid]:
                exclude_gids = comp_wind_dirs["upwind", gid]
                if downwind:
                    exclude_gids = np.append(
                        exclude_gids, comp_wind_dirs["downwind", gid]
                    )
                for n in exclude_gids:
                    check = comp_wind_dirs.exclude_sc_point_gid(n)
                    if check:
                        sc_gids = comp_wind_dirs[SupplyCurveField.SC_GID, n]
                        for sc_id in sc_gids:
                            if self._mask[sc_id]:
                                logger.debug(
                                    "Excluding sc_gid {}".format(sc_id)
                                )
                                self._mask[sc_id] = False

        return comp_wind_dirs

    @staticmethod
    def add_sum_cols(table, sum_cols):
        """Add a summation column to table.

        Parameters
        ----------
        table : pd.DataFrame
            Supply curve table.
        sum_cols : dict
            Mapping of new column label(s) to multiple column labels to sum.
            Example: sum_col={'total_cap_cost': ['cap_cost1', 'cap_cost2']}
            Which would add a new 'total_cap_cost' column which would be the
            sum of 'cap_cost1' and 'cap_cost2' if they are present in table.

        Returns
        -------
        table : pd.DataFrame
            Supply curve table with additional summation columns.
        """

        for new_label, sum_labels in sum_cols.items():
            missing = [s for s in sum_labels if s not in table]

            if any(missing):
                logger.info(
                    'Could not make sum column "{}", missing: {}'.format(
                        new_label, missing
                    )
                )
            else:
                sum_arr = np.zeros(len(table))
                for s in sum_labels:
                    temp = table[s].values
                    temp[np.isnan(temp)] = 0
                    sum_arr += temp

                table[new_label] = sum_arr

        return table

    def _full_sort(  # noqa: C901
        self,
        trans_table,
        trans_costs=None,
        avail_cap_frac=1,
        comp_wind_dirs=None,
        total_lcoe_fric=None,
        sort_on=SupplyCurveField.TOTAL_LCOE,
        columns=(
            SupplyCurveField.TRANS_GID,
            SupplyCurveField.TRANS_CAPACITY,
            SupplyCurveField.TRANS_TYPE,
            SupplyCurveField.TOTAL_TRANS_CAP_COST_PER_MW,
            SupplyCurveField.DIST_SPUR_KM,
            SupplyCurveField.LCOT,
            SupplyCurveField.TOTAL_LCOE,
        ),
        downwind=False,
    ):
        """
        Internal method to handle full supply curve sorting

        Parameters
        ----------
        trans_table : pandas.DataFrame
            Supply Curve Tranmission table to sort on
        trans_costs : str | dict, optional
            Transmission feature costs to use with TransmissionFeatures
            handler: line_tie_in_cost, line_cost, station_tie_in_cost,
            center_tie_in_cost, sink_tie_in_cost, by default None
        avail_cap_frac : int, optional
            Fraction of transmissions features capacity 'ac_cap' to make
            available for connection to supply curve points, by default 1
        comp_wind_dirs : CompetitiveWindFarms, optional
            Pre-initilized CompetitiveWindFarms instance, by default None
        total_lcoe_fric : ndarray, optional
            Vector of lcoe friction values, by default None
        sort_on : str, optional
            Column label to sort the Supply Curve table on. This affects the
            build priority - connections with the lowest value in this column
            will be built first, by default 'total_lcoe'
        columns : tuple, optional
            Columns to preserve in output connections dataframe,
            by default ('trans_gid', 'trans_capacity', 'trans_type',
                        'trans_cap_cost_per_mw', 'dist_km', 'lcot',
                        'total_lcoe')
        downwind : bool, optional
            Flag to remove downwind neighbors as well as upwind neighbors,
            by default False

        Returns
        -------
        supply_curve : pandas.DataFrame
            Updated sc_points table with transmission connections, LCOT
            and LCOE+LCOT based on full supply curve connections
        """
        trans_features = self._create_handler(
            self._trans_table,
            trans_costs=trans_costs,
            avail_cap_frac=avail_cap_frac,
        )
        init_list = [np.nan] * int(1 + np.max(self._sc_gids))
        columns = list(columns)
        if sort_on not in columns:
            columns.append(sort_on)

        conn_lists = {k: deepcopy(init_list) for k in columns}

        trans_sc_gids = trans_table[SupplyCurveField.SC_GID].values.astype(int)

        # syntax is final_key: source_key (source from trans_table)
        all_cols = list(columns)
        essentials = [SupplyCurveField.TRANS_GID,
                      SupplyCurveField.TRANS_CAPACITY,
                      SupplyCurveField.TRANS_TYPE,
                      SupplyCurveField.DIST_SPUR_KM,
                      SupplyCurveField.TOTAL_TRANS_CAP_COST_PER_MW,
                      SupplyCurveField.LCOT,
                      SupplyCurveField.TOTAL_LCOE]

        for col in essentials:
            if col not in all_cols:
                all_cols.append(col)

        arrays = {col: trans_table[col].values for col in all_cols}

        sc_capacities = trans_table[self._sc_capacity_col].values

        connected = 0
        progress = 0
        for i in range(len(trans_table)):
            sc_gid = trans_sc_gids[i]
            if self._mask[sc_gid]:
                connect = trans_features.connect(
                    arrays[SupplyCurveField.TRANS_GID][i], sc_capacities[i]
                )
                if connect:
                    connected += 1
                    logger.debug("Connecting sc gid {}".format(sc_gid))
                    self._mask[sc_gid] = False

                    for col_name, data_arr in arrays.items():
                        conn_lists[col_name][sc_gid] = data_arr[i]

                    if total_lcoe_fric is not None:
                        col_name = SupplyCurveField.TOTAL_LCOE_FRICTION
                        conn_lists[col_name][sc_gid] = total_lcoe_fric[i]

                    current_prog = connected // (len(self) / 100)
                    if current_prog > progress:
                        progress = current_prog
                        logger.info(
                            "{} % of supply curve points connected".format(
                                progress
                            )
                        )

                    if comp_wind_dirs is not None:
                        comp_wind_dirs = (
                            self._exclude_noncompetitive_wind_farms(
                                comp_wind_dirs, sc_gid, downwind=downwind
                            )
                        )

        index = range(0, int(1 + np.max(self._sc_gids)))
        connections = pd.DataFrame(conn_lists, index=index)
        connections.index.name = SupplyCurveField.SC_GID
        connections = connections.dropna(subset=[sort_on])
        connections = connections[columns].reset_index()

        sc_gids = self._sc_points[SupplyCurveField.SC_GID].values
        connected = connections[SupplyCurveField.SC_GID].values
        logger.debug('Connected gids {} out of total supply curve gids {}'
                     .format(len(connected), len(sc_gids)))
        unconnected = ~np.isin(sc_gids, connected)
        unconnected = sc_gids[unconnected].tolist()

        if unconnected:
            msg = (
                "{} supply curve points were not connected to tranmission! "
                "Unconnected sc_gid's: {}".format(
                    len(unconnected), unconnected
                )
            )
            logger.warning(msg)
            warn(msg)

        supply_curve = self._sc_points.merge(
            connections, on=SupplyCurveField.SC_GID)

        return supply_curve.reset_index(drop=True)

    def _check_feature_capacity(self, avail_cap_frac=1):
        """
        Add the transmission connection feature capacity to the trans table if
        needed
        """
        if SupplyCurveField.TRANS_CAPACITY not in self._trans_table:
            kwargs = {"avail_cap_frac": avail_cap_frac}
            fc = TF.feature_capacity(self._trans_table, **kwargs)
            self._trans_table = self._trans_table.merge(
                fc, on=SupplyCurveField.TRANS_GID)

    def _adjust_output_columns(self, columns, consider_friction):
        """Add extra output columns, if needed."""

        for col in _REQUIRED_COMPUTE_AND_OUTPUT_COLS:
            if col not in columns:
                columns.append(col)

        for col in _REQUIRED_OUTPUT_COLS:
            if col not in self._trans_table:
                self._trans_table[col] = np.nan
            if col not in columns:
                columns.append(col)

        missing_cols = [col for col in columns if col not in self._trans_table]
        if missing_cols:
            msg = (f"The following requested columns are not found in "
                   f"transmission table: {missing_cols}.\nSkipping...")
            logger.warning(msg)
            warn(msg)

        columns = [col for col in columns if col in self._trans_table]

        fric_col = SupplyCurveField.TOTAL_LCOE_FRICTION
        if consider_friction and fric_col in self._trans_table:
            columns.append(fric_col)

        return sorted(columns, key=_column_sort_key)

    def _determine_sort_on(self, sort_on):
        """Determine the `sort_on` column from user input and trans table"""
        r_cost_col = SupplyCurveField.REINFORCEMENT_COST_PER_MW
        found_reinforcement_costs = (
            r_cost_col in self._trans_table
            and not self._trans_table[r_cost_col].isna().all()
        )
        if found_reinforcement_costs:
            sort_on = sort_on or "lcoe_no_reinforcement"
        return sort_on or SupplyCurveField.TOTAL_LCOE

    def full_sort(
        self,
        fcr,
        transmission_costs=None,
        avail_cap_frac=1,
        line_limited=False,
        connectable=True,
        max_workers=None,
        consider_friction=True,
        sort_on=None,
        columns=(
            SupplyCurveField.TRANS_GID,
            SupplyCurveField.TRANS_CAPACITY,
            SupplyCurveField.TRANS_TYPE,
            SupplyCurveField.TOTAL_TRANS_CAP_COST_PER_MW,
            SupplyCurveField.DIST_SPUR_KM,
            SupplyCurveField.LCOT,
            SupplyCurveField.TOTAL_LCOE,
        ),
        wind_dirs=None,
        n_dirs=2,
        downwind=False,
        offshore_compete=False,
    ):
        """
        run full supply curve sorting

        Parameters
        ----------
        fcr : float
            Fixed charge rate, used to compute LCOT
        transmission_costs : str | dict, optional
            Transmission feature costs to use with TransmissionFeatures
            handler: line_tie_in_cost, line_cost, station_tie_in_cost,
            center_tie_in_cost, sink_tie_in_cost, by default None
        avail_cap_frac : int, optional
            Fraction of transmissions features capacity 'ac_cap' to make
            available for connection to supply curve points, by default 1
        line_limited : bool, optional
            Flag to have substation connection is limited by maximum capacity
            of the attached lines, legacy method, by default False
        connectable : bool, optional
            Flag to only compute tranmission capital cost if transmission
            feature has enough available capacity, by default True
        max_workers : int | NoneType, optional
            Number of workers to use to compute lcot, if > 1 run in parallel.
            None uses all available cpu's. by default None
        consider_friction : bool, optional
            Flag to consider friction layer on LCOE when "mean_lcoe_friction"
            is in the sc points input, by default True
        sort_on : str, optional
            Column label to sort the Supply Curve table on. This affects the
            build priority - connections with the lowest value in this column
            will be built first, by default `None`, which will use
            total LCOE without any reinforcement costs as the sort value.
        columns : list | tuple, optional
            Columns to preserve in output connections dataframe,
            by default ('trans_gid', 'trans_capacity', 'trans_type',
            'trans_cap_cost_per_mw', 'dist_km', 'lcot', 'total_lcoe')
        wind_dirs : pandas.DataFrame | str, optional
            path to .csv or reVX.wind_dirs.wind_dirs.WindDirs output with
            the neighboring supply curve point gids and power-rose value at
            each cardinal direction, by default None
        n_dirs : int, optional
            Number of prominent directions to use, by default 2
        downwind : bool, optional
            Flag to remove downwind neighbors as well as upwind neighbors,
            by default False
        offshore_compete : bool, default
            Flag as to whether offshore farms should be included during
            CompetitiveWindFarms, by default False

        Returns
        -------
        supply_curve : pandas.DataFrame
            Updated sc_points table with transmission connections, LCOT
            and LCOE+LCOT based on full supply curve connections
        """
        logger.info("Starting full competitive supply curve sort.")
        self._check_substation_conns(self._trans_table)
        self.compute_total_lcoe(
            fcr,
            transmission_costs=transmission_costs,
            avail_cap_frac=avail_cap_frac,
            line_limited=line_limited,
            connectable=connectable,
            max_workers=max_workers,
            consider_friction=consider_friction,
        )
        self._check_feature_capacity(avail_cap_frac=avail_cap_frac)

        if isinstance(columns, tuple):
            columns = list(columns)

        columns = self._adjust_output_columns(columns, consider_friction)
        sort_on = self._determine_sort_on(sort_on)

        trans_table = self._trans_table.copy()
        pos = trans_table[SupplyCurveField.LCOT].isnull()
        trans_table = trans_table.loc[~pos].sort_values(
            [sort_on, SupplyCurveField.TRANS_GID]
        )

        total_lcoe_fric = None
        col_in_table = SupplyCurveField.MEAN_LCOE_FRICTION in trans_table
        if consider_friction and col_in_table:
            total_lcoe_fric = \
                trans_table[SupplyCurveField.TOTAL_LCOE_FRICTION].values

        comp_wind_dirs = None
        if wind_dirs is not None:
            msg = "Excluding {} upwind".format(n_dirs)
            if downwind:
                msg += " and downwind"

            msg += " onshore"
            if offshore_compete:
                msg += " and offshore"

            msg += " windfarms"
            logger.info(msg)
            comp_wind_dirs = CompetitiveWindFarms(
                wind_dirs,
                self._sc_points,
                n_dirs=n_dirs,
                offshore=offshore_compete,
            )

        supply_curve = self._full_sort(
            trans_table,
            trans_costs=transmission_costs,
            avail_cap_frac=avail_cap_frac,
            comp_wind_dirs=comp_wind_dirs,
            total_lcoe_fric=total_lcoe_fric,
            sort_on=sort_on,
            columns=columns,
            downwind=downwind,
        )

        return supply_curve

    def simple_sort(
        self,
        fcr,
        transmission_costs=None,
        avail_cap_frac=1,
        max_workers=None,
        consider_friction=True,
        sort_on=None,
        columns=DEFAULT_COLUMNS,
        wind_dirs=None,
        n_dirs=2,
        downwind=False,
        offshore_compete=False,
    ):
        """
        Run simple supply curve sorting that does not take into account
        available capacity

        Parameters
        ----------
        fcr : float
            Fixed charge rate, used to compute LCOT
        transmission_costs : str | dict, optional
            Transmission feature costs to use with TransmissionFeatures
            handler: line_tie_in_cost, line_cost, station_tie_in_cost,
            center_tie_in_cost, sink_tie_in_cost, by default None
        avail_cap_frac : int, optional
            Fraction of transmissions features capacity 'ac_cap' to make
            available for connection to supply curve points, by default 1
        line_limited : bool, optional
            Flag to have substation connection is limited by maximum capacity
            of the attached lines, legacy method, by default False
        connectable : bool, optional
            Flag to only compute tranmission capital cost if transmission
            feature has enough available capacity, by default True
        max_workers : int | NoneType, optional
            Number of workers to use to compute lcot, if > 1 run in parallel.
            None uses all available cpu's. by default None
        consider_friction : bool, optional
            Flag to consider friction layer on LCOE when "mean_lcoe_friction"
            is in the sc points input, by default True
        sort_on : str, optional
            Column label to sort the Supply Curve table on. This affects the
            build priority - connections with the lowest value in this column
            will be built first, by default `None`, which will use
            total LCOE without any reinforcement costs as the sort value.
        columns : list | tuple, optional
            Columns to preserve in output connections dataframe.
            By default, :obj:`DEFAULT_COLUMNS`.
        wind_dirs : pandas.DataFrame | str, optional
            path to .csv or reVX.wind_dirs.wind_dirs.WindDirs output with
            the neighboring supply curve point gids and power-rose value at
            each cardinal direction, by default None
        n_dirs : int, optional
            Number of prominent directions to use, by default 2
        downwind : bool, optional
            Flag to remove downwind neighbors as well as upwind neighbors
        offshore_compete : bool, default
            Flag as to whether offshore farms should be included during
            CompetitiveWindFarms, by default False

        Returns
        -------
        supply_curve : pandas.DataFrame
            Updated sc_points table with transmission connections, LCOT
            and LCOE+LCOT based on simple supply curve connections
        """
        logger.info("Starting simple supply curve sort (no capacity limits).")
        self.compute_total_lcoe(
            fcr,
            transmission_costs=transmission_costs,
            avail_cap_frac=avail_cap_frac,
            connectable=False,
            max_workers=max_workers,
            consider_friction=consider_friction,
        )
        sort_on = self._determine_sort_on(sort_on)

        if isinstance(columns, tuple):
            columns = list(columns)
        columns = self._adjust_output_columns(columns, consider_friction)

        trans_table = self._trans_table.copy()
        connections = trans_table.sort_values(
            [sort_on, SupplyCurveField.TRANS_GID])
        connections = connections.groupby(SupplyCurveField.SC_GID).first()
        connections = connections[columns].reset_index()

        supply_curve = self._sc_points.merge(connections,
                                             on=SupplyCurveField.SC_GID)
        if wind_dirs is not None:
            supply_curve = CompetitiveWindFarms.run(
                wind_dirs,
                supply_curve,
                n_dirs=n_dirs,
                offshore=offshore_compete,
                sort_on=sort_on,
                downwind=downwind,
            )

        supply_curve = supply_curve.reset_index(drop=True)

        return supply_curve

    def run(
        self,
        out_fpath,
        fixed_charge_rate,
        simple=True,
        avail_cap_frac=1,
        line_limited=False,
        transmission_costs=None,
        consider_friction=True,
        sort_on=None,
        columns=DEFAULT_COLUMNS,
        max_workers=None,
        competition=None,
    ):
        """Run Supply Curve Transmission calculations.

        Run full supply curve taking into account available capacity of
        tranmission features when making connections.

        Parameters
        ----------
        out_fpath : str
            Full path to output CSV file. Does not need to include file
            ending - it will be added automatically if missing.
        fixed_charge_rate : float
            Fixed charge rate, (in decimal form: 5% = 0.05). This value
            is used to compute LCOT.
        simple : bool, optional
            Option to run the simple sort (does not keep track of
            capacity available on the existing transmission grid). If
            ``False``, a full transmission sort (where connections are
            limited based on available transmission capacity) is run.
            Note that the full transmission sort requires the
            `avail_cap_frac` and `line_limited` inputs.
            By default, ``True``.
        avail_cap_frac : int, optional
            This input has no effect if ``simple=True``. Fraction of
            transmissions features capacity ``ac_cap`` to make available
            for connection to supply curve points. By default, ``1``.
        line_limited : bool, optional
            This input has no effect if ``simple=True``. Flag to have
            substation connection limited by maximum capacity
            of the attached lines. This is a legacy method.
            By default, ``False``.
        transmission_costs : str | dict, optional
            Dictionary of transmission feature costs or path to JSON
            file containing a dictionary of transmission feature costs.
            These costs are used to compute transmission capital cost
            if the input transmission tables do not have a
            ``"trans_cap_cost"`` column (this input is ignored
            otherwise). The dictionary must include:

                - line_tie_in_cost
                - line_cost
                - station_tie_in_cost
                - center_tie_in_cost
                - sink_tie_in_cost

            By default, ``None``.
        consider_friction : bool, optional
            Flag to add a new ``"total_lcoe_friction"`` column to the
            supply curve output that contains the sum of the computed
            ``"total_lcoe"`` value and the input
            ``"mean_lcoe_friction"`` values. If ``"mean_lcoe_friction"``
            is not in the `sc_points` input, this option is ignored.
            By default, ``True``.
        sort_on : str, optional
            Column label to sort the supply curve table on. This affects
            the build priority when doing a "full" sort - connections
            with the lowest value in this column will be built first.
            For a "simple" sort, only connections with the lowest value
            in this column will be considered. If ``None``, the sort is
            performed on the total LCOE *without* any reinforcement
            costs added (this is typically what you want - it avoids
            unrealistically long spur-line connections).
            By default ``None``.
        columns : list | tuple, optional
            Columns to preserve in output supply curve dataframe.
            By default, :obj:`DEFAULT_COLUMNS`.
        max_workers : int, optional
            Number of workers to use to compute LCOT. If > 1,
            computation is run in parallel. If ``None``, computation
            uses all available CPU's. By default, ``None``.
        competition : dict, optional
            Optional dictionary of arguments for competitive wind farm
            exclusions, which removes supply curve points upwind (and
            optionally downwind) of the lowest LCOE supply curves.
            If ``None``, no competition is applied. Otherwise, this
            dictionary can have up to four keys:

                - ``wind_dirs`` (required) : A path to a CSV file or
                  :py:class:`reVX ProminentWindDirections
                  <reVX.wind_dirs.prominent_wind_dirs.ProminentWindDirections>`
                  output with the neighboring supply curve point gids
                  and power-rose values at each cardinal direction.
                - ``n_dirs`` (optional) : An integer representing the
                  number of prominent directions to use during wind farm
                  competition. By default, ``2``.
                - ``downwind`` (optional) : A flag indicating that
                  downwind neighbors should be removed in addition to
                  upwind neighbors during wind farm competition.
                  By default, ``False``.
                - ``offshore_compete`` (optional) : A flag indicating
                  that offshore farms should be included during wind
                  farm competition. By default, ``False``.

            By default ``None``.

        Returns
        -------
        str
            Path to output supply curve.
        """
        kwargs = {
            "fcr": fixed_charge_rate,
            "transmission_costs": transmission_costs,
            "consider_friction": consider_friction,
            "sort_on": sort_on,
            "columns": columns,
            "max_workers": max_workers,
        }
        kwargs.update(competition or {})

        if simple:
            supply_curve = self.simple_sort(**kwargs)
        else:
            kwargs["avail_cap_frac"] = avail_cap_frac
            kwargs["line_limited"] = line_limited
            supply_curve = self.full_sort(**kwargs)

        out_fpath = _format_sc_out_fpath(out_fpath)
        supply_curve.to_csv(out_fpath, index=False)

        return out_fpath


def _format_sc_out_fpath(out_fpath):
    """Add CSV file ending and replace underscore, if necessary."""
    if not out_fpath.endswith(".csv"):
        out_fpath = "{}.csv".format(out_fpath)

    project_dir, out_fn = os.path.split(out_fpath)
    out_fn = out_fn.replace("supply_curve", "supply-curve")
    return os.path.join(project_dir, out_fn)


def _column_sort_key(col):
    """Determine the sort order of the input column. """
    col_value = _REQUIRED_COMPUTE_AND_OUTPUT_COLS.get(col)
    if col_value is None:
        col_value = _REQUIRED_OUTPUT_COLS.get(col)
    if col_value is None:
        col_value = 1e6

    return col_value, str(col)



# -*- coding: utf-8 -*-
"""Collection of functions used to hybridize columns in rep profiles meta.

@author: ppinchuk
"""
from reV.utilities import SupplyCurveField


def aggregate_solar_capacity(h):
    """Compute the total solar capcity allowed in hybridization.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the capacity allowed in the hybrid
        capacity sum, or `None` if 'hybrid_solar_capacity' already
        exists.

    Notes
    -----
    No limiting is done on the ratio of wind to solar. This method
    checks for an existing 'hybrid_solar_capacity'. If one does not
    exist, it is assumed that there is no limit on the solar to wind
    capacity ratio and the solar capacity is copied into this new
    column.
    """
    if f'hybrid_solar_{SupplyCurveField.CAPACITY_AC_MW}' in h.hybrid_meta:
        return None
    return h.hybrid_meta[f'solar_{SupplyCurveField.CAPACITY_AC_MW}']


def aggregate_wind_capacity(h):
    """Compute the total wind capcity allowed in hybridization.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the capacity allowed in the hybrid
        capacity sum, or `None` if 'hybrid_solar_capacity' already
        exists.

    Notes
    -----
    No limiting is done on the ratio of wind to solar. This method
    checks for an existing 'hybrid_wind_capacity'. If one does not
    exist, it is assumed that there is no limit on the solar to wind
    capacity ratio and the wind capacity is copied into this new column.
    """
    if f'hybrid_wind_{SupplyCurveField.CAPACITY_AC_MW}' in h.hybrid_meta:
        return None
    return h.hybrid_meta[f'wind_{SupplyCurveField.CAPACITY_AC_MW}']


def aggregate_capacity(h):
    """Compute the total capcity by summing the individual capacities.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the aggregated capacity, or `None`
        if the capacity columns are missing.
    """
    sc = f'hybrid_solar_{SupplyCurveField.CAPACITY_AC_MW}'
    wc = f'hybrid_wind_{SupplyCurveField.CAPACITY_AC_MW}'
    missing_solar_cap = sc not in h.hybrid_meta.columns
    missing_wind_cap = wc not in h.hybrid_meta.columns
    if missing_solar_cap or missing_wind_cap:
        return None

    total_cap = h.hybrid_meta[sc] + h.hybrid_meta[wc]
    return total_cap


def aggregate_capacity_factor(h):
    """Compute the capacity-weighted mean capcity factor.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the aggregated capacity, or `None`
        if the capacity and/or mean_cf columns are missing.
    """

    sc = f'hybrid_solar_{SupplyCurveField.CAPACITY_AC_MW}'
    wc = f'hybrid_wind_{SupplyCurveField.CAPACITY_AC_MW}'
    scf = f'solar_{SupplyCurveField.MEAN_CF_AC}'
    wcf = f'wind_{SupplyCurveField.MEAN_CF_AC}'
    missing_solar_cap = sc not in h.hybrid_meta.columns
    missing_wind_cap = wc not in h.hybrid_meta.columns
    missing_solar_mean_cf = scf not in h.hybrid_meta.columns
    missing_wind_mean_cf = wcf not in h.hybrid_meta.columns
    missing_any = (missing_solar_cap or missing_wind_cap
                   or missing_solar_mean_cf or missing_wind_mean_cf)
    if missing_any:
        return None

    solar_cf_weighted = h.hybrid_meta[sc] * h.hybrid_meta[scf]
    wind_cf_weighted = h.hybrid_meta[wc] * h.hybrid_meta[wcf]
    total_capacity = aggregate_capacity(h)
    hybrid_cf = (solar_cf_weighted + wind_cf_weighted) / total_capacity
    return hybrid_cf


HYBRID_METHODS = {
    f'hybrid_solar_{SupplyCurveField.CAPACITY_AC_MW}': (
        aggregate_solar_capacity
    ),
    f'hybrid_wind_{SupplyCurveField.CAPACITY_AC_MW}': aggregate_wind_capacity,
    f'hybrid_{SupplyCurveField.CAPACITY_AC_MW}': aggregate_capacity,
    f'hybrid_{SupplyCurveField.MEAN_CF_AC}': aggregate_capacity_factor
}



# -*- coding: utf-8 -*-
"""
Generation CLI utility functions.
"""
import pprint
import logging

from gaps.cli import as_click_command, CLICommandFromClass

from reV.generation.generation import Gen
from reV.utilities import ModuleName
from reV.utilities.cli_functions import format_analysis_years, init_cli_logging
from reV.utilities.exceptions import ConfigError


logger = logging.getLogger(__name__)


def _preprocessor(config, job_name, log_directory, verbose,
                  analysis_years=None):
    """Preprocess generation config user input.

    Parameters
    ----------
    config : dict
        User configuration file input as (nested) dict.
    job_name : str
        Name of ``reV`` job being run.
    log_directory : str
        Path to log output directory.
    verbose : bool
        Flag to signal ``DEBUG`` verbosity (``verbose=True``).
    analysis_years : int | list, optional
        A single year or list of years to perform analysis for. These
        years will be used to fill in any brackets ``{}`` in the
        `resource_file` input. If ``None``, the `resource_file`
        input is assumed to be the full path to the single resource
        file to be processed.  By default, ``None``.

    Returns
    -------
    dict
        Updated config file.
    """
    init_cli_logging(job_name, log_directory, verbose)
    config.get("execution_control", {}).setdefault("max_workers")
    analysis_years = format_analysis_years(analysis_years)

    config["resource_file"] = _parse_res_files(config["resource_file"],
                                               analysis_years)
    lr_res_file = config.get("low_res_resource_file")
    if lr_res_file is None:
        config["low_res_resource_file"] = [None] * len(analysis_years)
    else:
        config["low_res_resource_file"] = _parse_res_files(lr_res_file,
                                                           analysis_years)

    config['technology'] = (config['technology'].lower()
                            .replace(' ', '').replace('_', ''))
    _log_generation_cli_inputs(config)
    return config


def _parse_res_files(res_fps, analysis_years):
    """Parse the base resource file input into correct ordered list format
    with year imputed in the {} format string"""

    # get base filename, may have {} for year format
    if isinstance(res_fps, str) and '{}' in res_fps:
        # need to make list of res files for each year
        res_fps = [res_fps.format(year) for year in analysis_years]
    elif isinstance(res_fps, str):
        # only one resource file request, still put in list
        res_fps = [res_fps]
    elif not isinstance(res_fps, (list, tuple)):
        msg = ('Bad "resource_file" type, needed str, list, or tuple '
               'but received: {}, {}'
               .format(res_fps, type(res_fps)))
        logger.error(msg)
        raise ConfigError(msg)

    if len(res_fps) != len(analysis_years):
        msg = ('The number of resource files does not match '
               'the number of analysis years!'
               '\n\tResource files: \n\t\t{}'
               '\n\tYears: \n\t\t{}'
               .format(res_fps, analysis_years))
        logger.error(msg)
        raise ConfigError(msg)

    return res_fps


def _log_generation_cli_inputs(config):
    """Log initial generation CLI inputs"""

    logger.info('The following project points were specified: "{}"'
                .format(config.get('project_points', None)))
    logger.info('The following SAM configs are available to this run:\n{}'
                .format(pprint.pformat(config.get('sam_files', None),
                                       indent=4)))
    logger.info('The following is being used for site specific input data: '
                '"{}"'.format(config.get("site_data")))


gen_command = CLICommandFromClass(Gen, method="run",
                                  name=str(ModuleName.GENERATION),
                                  add_collect=False,
                                  split_keys=["project_points",
                                              ("resource_file",
                                               "low_res_resource_file")],
                                  config_preprocessor=_preprocessor)
main = as_click_command(gen_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Generation CLI.')
        raise



# -*- coding: utf-8 -*-
"""
reV Project Points Configuration
"""

import copy
import logging
import os
from warnings import warn

import numpy as np
import pandas as pd
from rex.multi_file_resource import MultiFileResource
from rex.resource import Resource
from rex.resource_extraction.resource_extraction import (
    MultiFileResourceX,
    ResourceX,
)
from rex.utilities import check_res_file, parse_table

from reV.config.curtailment import Curtailment
from reV.config.sam_config import SAMConfig
from reV.utilities import SiteDataField, SupplyCurveField
from reV.utilities.exceptions import ConfigError, ConfigWarning

logger = logging.getLogger(__name__)


class PointsControl:
    """Class to manage and split ProjectPoints."""

    def __init__(self, project_points, sites_per_split=100):
        """
        Parameters
        ----------
        project_points : reV.config.ProjectPoints
            ProjectPoints instance to be split between execution workers.
        sites_per_split : int
            Sites per project points split instance returned in the __next__
            iterator function.
        """

        self._project_points = project_points
        self._sites_per_split = sites_per_split
        self._split_range = []
        self._i = 0
        self._iter_list = []

    def __iter__(self):
        """Initialize the iterator by pre-splitting into a list attribute."""
        last_site = 0
        ilim = len(self.project_points)

        logger.debug(
            "PointsControl iterator initializing with sites "
            "{} through {}".format(
                self.project_points.sites[0], self.project_points.sites[-1]
            )
        )

        # pre-initialize all iter objects
        while True:
            i0 = last_site
            i1 = np.min([i0 + self.sites_per_split, ilim])
            if i0 == i1:
                break

            last_site = i1

            new = self.split(
                i0,
                i1,
                self.project_points,
                sites_per_split=self.sites_per_split,
            )
            new._split_range = [i0, i1]
            self._iter_list.append(new)

        logger.debug(
            "PointsControl stopped iteration at attempted "
            "index of {}. Length of iterator is: {}".format(i1, len(self))
        )
        return self

    def __next__(self):
        """Iterate through and return next site resource data.

        Returns
        -------
        next_pc : config.PointsControl
            Split instance of this class with a subset of project points based
            on the number of sites per split.
        """
        if self._i < self.N:
            # Get next PointsControl from the iter list
            next_pc = self._iter_list[self._i]
        else:
            # No more points controllers left in initialized list
            raise StopIteration

        logger.debug(
            "PointsControl passing site project points "
            "with indices {} to {} on iteration #{} ".format(
                next_pc.split_range[0], next_pc.split_range[1], self._i
            )
        )
        self._i += 1
        return next_pc

    def __repr__(self):
        msg = "{} with {} sites from gid {} through {}".format(
            self.__class__.__name__,
            len(self.project_points),
            self.sites[0],
            self.sites[-1],
        )
        return msg

    def __len__(self):
        """Len is the number of possible iterations aka splits."""
        return int(np.ceil(len(self.project_points) / self.sites_per_split))

    @property
    def N(self):
        """
        Length of current iterator list

        Returns
        -------
        N : int
            Number of iterators in list
        """
        return len(self._iter_list)

    @property
    def sites_per_split(self):
        """Get the iterator increment: number of sites per split.

        Returns
        -------
        _sites_per_split : int
            Sites per split iter object.
        """
        return self._sites_per_split

    @property
    def project_points(self):
        """Get the project points property.

        Returns
        -------
        _project_points : reV.config.project_points.ProjectPoints
            ProjectPoints instance corresponding to this PointsControl
            instance.
        """
        return self._project_points

    @property
    def sites(self):
        """Get the project points sites for this instance.

        Returns
        -------
        sites : list
            List of sites belonging to the _project_points attribute.
        """
        return self._project_points.sites

    @property
    def split_range(self):
        """Get the current split range property.

        Returns
        -------
        _split_range : list
            Two-entry list that indicates the starting and finishing
            (inclusive, exclusive, respectively) indices of a split instance
            of the PointsControl object. This is set in the iterator dunder
            methods of PointsControl.
        """
        return self._split_range

    @classmethod
    def split(cls, i0, i1, project_points, sites_per_split=100):
        """Split this execution by splitting the project points attribute.

        Parameters
        ----------
        i0/i1 : int
            Beginning/end (inclusive/exclusive, respectively) index split
            parameters for ProjectPoints.split() method.
        project_points : reV.config.ProjectPoints
            Project points instance that will be split.
        sites_per_split : int
            Sites per project points split instance returned in the __next__
            iterator function.

        Returns
        -------
        sub : PointsControl
            New instance of PointsControl with a subset of the original
            project points.
        """
        i0 = int(i0)
        i1 = int(i1)
        new_points = ProjectPoints.split(i0, i1, project_points)
        sub = cls(new_points, sites_per_split=sites_per_split)
        return sub


class ProjectPoints:
    """Class to manage site and SAM input configuration requests.

    Examples
    --------
    >>> import os
    >>> from reV import TESTDATADIR
    >>> from reV.config.project_points import ProjectPoints
    >>>
    >>> points = slice(0, 100)
    >>> sam_file = os.path.join(TESTDATADIR, 'SAM/naris_pv_1axis_inv13.json')
    >>> pp = ProjectPoints(points, sam_file)
    >>>
    >>> config_id_site0, SAM_config_dict_site0 = pp[0]
    >>> site_list_or_slice = pp.sites
    >>> site_list_or_slice = pp.get_sites_from_config(config_id)
    >>> ProjectPoints_sub = pp.split(0, 10, project_points)
    >>> h_list = pp.h
    """

    def __init__(
        self, points, sam_configs, tech=None, res_file=None, curtailment=None
    ):
        """
        Parameters
        ----------
        points : int | slice | list | tuple | str | pd.DataFrame | dict
            Slice specifying project points, string pointing to a project
            points csv, or a dataframe containing the effective csv contents.
            Can also be a single integer site value.
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.
        tech : str, optional
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed,
            by default None
        res_file : str | NoneType
            Optional resource file to find maximum length of project points if
            points slice stop is None.
        curtailment : NoneType | dict | str | config.curtailment.Curtailment
            Inputs for curtailment parameters. If not None, curtailment inputs
            are expected. Can be:

                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config json file with path (str)
                - Instance of curtailment config object
                  (config.curtailment.Curtailment)

        """
        # set protected attributes
        self._df = self._parse_points(points, res_file=res_file)
        self._sam_config_obj = self._parse_sam_config(sam_configs)
        self._check_points_config_mapping()
        self._tech = str(tech)
        self._h = self._d = None
        self._curtailment = self._parse_curtailment(curtailment)

    def __getitem__(self, site):
        """Get the SAM config ID and dictionary for the requested site.

        Parameters
        ----------
        site : int | str
            Site number (gid) of interest (typically the resource gid).

        Returns
        -------
        config_id : str
            Configuration ID (variable name) specified in the sam_generation
            config section.
        config : dict
            Actual SAM input values in a single level dictionary with variable
            names (keys) and values.
        """

        site_bool = self.df[SiteDataField.GID] == site
        try:
            config_id = self.df.loc[site_bool, SiteDataField.CONFIG].values[0]
        except (KeyError, IndexError) as ex:
            msg = (
                "Site {} not found in this instance of "
                "ProjectPoints. Available sites include: {}".format(
                    site, self.sites
                )
            )
            logger.exception(msg)
            raise KeyError(msg) from ex

        return config_id, copy.deepcopy(self.sam_inputs[config_id])

    def __repr__(self):
        msg = "{} with {} sites from gid {} through {}".format(
            self.__class__.__name__, len(self), self.sites[0], self.sites[-1]
        )
        return msg

    def __len__(self):
        """Length of this object is the number of sites."""
        return len(self.sites)

    @property
    def df(self):
        """Get the project points dataframe property.

        Returns
        -------
        _df : pd.DataFrame
            Table of sites and corresponding SAM configuration IDs.
            Has columns "gid" and 'config'.
        """
        return self._df

    @property
    def sam_config_ids(self):
        """Get the SAM configs dictionary property.

        Returns
        -------
        dict
            Multi-level dictionary containing multiple SAM input config files.
            The top level key is the SAM config ID, top level value is the SAM
            config file path
        """
        return sorted(self._sam_config_obj)

    @property
    def sam_config_obj(self):
        """Get the SAM config object.

        Returns
        -------
        _sam_config_obj : reV.config.sam_config.SAMConfig
            SAM configuration object.
        """
        return self._sam_config_obj

    @property
    def sam_inputs(self):
        """Get the SAM configuration inputs dictionary property.

        Returns
        -------
        dict
            Multi-level dictionary containing multiple SAM input
            configurations. The top level key is the SAM config ID, top level
            value is the SAM config. Each SAM config is a dictionary with keys
            equal to input names, values equal to the actual inputs.
        """
        return self.sam_config_obj.inputs

    @property
    def all_sam_input_keys(self):
        """Get a list of unique input keys from all SAM technology configs.

        Returns
        -------
        all_sam_input_keys : list
            List of unique strings where each string is a input key for the
            SAM technology configs. For example, "gcr" or "losses" for PVWatts
            or "wind_turbine_hub_ht" for windpower.
        """
        keys = []
        for sam_config in self.sam_inputs.values():
            keys += list(sam_config.keys())

        keys = list(set(keys))

        return keys

    @property
    def gids(self):
        """Get the list of gids (resource file index values) belonging to this
        instance of ProjectPoints. This is an alias of self.sites.

        Returns
        -------
        gids : list
            List of integer gids (resource file index values) belonging to this
            instance of ProjectPoints. This is an alias of self.sites.
        """
        return self.sites

    @property
    def sites(self):
        """Get the list of sites (resource file gids) belonging to this
        instance of ProjectPoints.

        Returns
        -------
        sites : list
            List of integer sites (resource file gids) belonging to this
            instance of ProjectPoints.
        """
        return self.df[SiteDataField.GID].values.tolist()

    @property
    def sites_as_slice(self):
        """Get the sites in slice format.

        Returns
        -------
        sites_as_slice : list | slice
            Sites slice belonging to this instance of ProjectPoints.
            The type is slice if possible. Will be a list only if sites are
            non-sequential.
        """
        # try_slice is what the sites list would be if it is sequential
        if len(self.sites) > 1:
            try_step = self.sites[1] - self.sites[0]
        else:
            try_step = 1
        try_slice = slice(self.sites[0], self.sites[-1] + 1, try_step)
        try_list = list(range(*try_slice.indices(try_slice.stop)))

        if self.sites == try_list:
            # try_slice is equivelant to the site list
            sites_as_slice = try_slice
        else:
            # cannot be converted to a sequential slice, return list
            sites_as_slice = self.sites

        return sites_as_slice

    @property
    def tech(self):
        """Get the tech property from the config.

        Returns
        -------
        _tech : str
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed.
        """
        return "windpower" if "wind" in self._tech.lower() else self._tech

    @property
    def h(self):
        """Get the hub heights corresponding to the site list.

        Returns
        -------
        _h : list | NoneType
            Hub heights corresponding to each site, taken from the sam config
            for each site. This is None if the technology is not wind.
        """
        h_var = "wind_turbine_hub_ht"
        if self._h is None:
            if "wind" in self.tech:
                # wind technology, get a list of h values
                self._h = [self[site][1][h_var] for site in self.sites]

        return self._h

    @property
    def d(self):
        """Get the depths (m) corresponding to the site list.

        Returns
        -------
        _d : list | NoneType
            Resource depths (m) corresponding to each site, taken from
            the sam config for each site. This is None if the technology
            is not geothermal.
        """
        d_var = "resource_depth"
        if self._d is None:
            if "geothermal" in self.tech:
                if d_var in self.df:
                    self._d = list(self.df[d_var])
                else:
                    self._d = [self[site][1][d_var] for site in self.sites]

        return self._d

    @property
    def curtailment(self):
        """Get the curtailment config object.

        Returns
        -------
        _curtailment : NoneType | reV.config.curtailment.Curtailment
            None if no curtailment, reV curtailment config object if
            curtailment is being assessed.
        """
        return self._curtailment

    @staticmethod
    def _parse_csv(fname):
        """Import project points from .csv

        Parameters
        ----------
        fname : str
            Project points .csv file (with path). Must have 'gid' and
            'config' column names.

        Returns
        -------
        df : pd.DataFrame
            DataFrame mapping sites (gids) to SAM technology (config)
        """
        fname = fname.strip()
        if fname.endswith(".csv"):
            df = pd.read_csv(fname)
        else:
            raise ValueError(
                "Config project points file must be "
                ".csv, but received: {}".format(fname)
            )

        return df

    @staticmethod
    def _parse_sites(points, res_file=None):
        """Parse project points from list or slice

        Parameters
        ----------
        points : int | str | pd.DataFrame | slice | list
            Slice specifying project points, string pointing to a project
            points csv, or a dataframe containing the effective csv contents.
            Can also be a single integer site value.
        res_file : str | NoneType
            Optional resource file to find maximum length of project points if
            points slice stop is None.

        Returns
        -------
        df : pd.DataFrame
            DataFrame mapping sites (gids) to SAM technology (config)
        """
        df = pd.DataFrame(columns=[SiteDataField.GID, SiteDataField.CONFIG])
        if isinstance(points, int):
            points = [points]
        if isinstance(points, (list, tuple, np.ndarray)):
            # explicit site list, set directly
            if any(isinstance(i, (list, tuple, np.ndarray)) for i in points):
                msg = "Provided project points is not flat: {}!".format(points)
                logger.error(msg)
                raise RuntimeError(msg)

            df[SiteDataField.GID] = points
        elif isinstance(points, slice):
            stop = points.stop
            if stop is None:
                if res_file is None:
                    raise ValueError(
                        "Must supply a resource file if "
                        "points is a slice of type "
                        " slice(*, None, *)"
                    )

                multi_h5_res, _ = check_res_file(res_file)
                if multi_h5_res:
                    stop = MultiFileResource(res_file).shape[1]
                else:
                    stop = Resource(res_file).shape[1]

            df[SiteDataField.GID] = list(range(*points.indices(stop)))
        else:
            raise TypeError(
                "Project Points sites needs to be set as a list, "
                "tuple, or slice, but was set as: {}".format(type(points))
            )

        df[SiteDataField.CONFIG] = None

        return df

    @classmethod
    def _parse_points(cls, points, res_file=None):
        """Generate the project points df from inputs

        Parameters
        ----------
        points : int | str | pd.DataFrame | slice | list | dict
            Slice specifying project points, string pointing to a project
            points csv, or a dataframe containing the effective csv contents.
            Can also be a single integer site value.
        res_file : str | NoneType
            Optional resource file to find maximum length of project points if
            points slice stop is None.

        Returns
        -------
        df : pd.DataFrame
            DataFrame mapping sites (gids) to SAM technology (config)
        """
        if isinstance(points, str):
            df = cls._parse_csv(points)
        elif isinstance(points, dict):
            df = pd.DataFrame(points)
        elif isinstance(points, (int, slice, list, tuple, np.ndarray)):
            df = cls._parse_sites(points, res_file=res_file)
        elif isinstance(points, pd.DataFrame):
            df = points
        else:
            raise ValueError(
                "Cannot parse Project points data from {}".format(type(points))
            )
        df = df.rename(SupplyCurveField.map_to(SiteDataField), axis=1)
        if SiteDataField.GID not in df.columns:
            raise KeyError(
                "Project points data must contain "
                f"{SiteDataField.GID} column."
            )

        # pylint: disable=no-member
        if SiteDataField.CONFIG not in df.columns:
            df[SiteDataField.CONFIG] = None

        gids = df[SiteDataField.GID].values
        if not np.array_equal(np.sort(gids), gids):
            msg = (
                "WARNING: points are not in sequential order and will be "
                "sorted! The original order is being preserved under "
                'column "points_order"'
            )
            logger.warning(msg)
            warn(msg)
            df["points_order"] = df.index.values
            df = df.sort_values(SiteDataField.GID).reset_index(drop=True)

        return df

    @staticmethod
    def _parse_sam_config(sam_config):
        """
        Create SAM files dictionary.

        Parameters
        ----------
        sam_config : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s) or SAM config
            dict(s). Keys are the SAM config ID(s). Can also be a single
            config file str. Can also be a pre loaded SAMConfig object.

        Returns
        -------
        _sam_config_obj : reV.config.sam_config.SAMConfig
            SAM configuration object.
        """

        if isinstance(sam_config, SAMConfig):
            return sam_config

        if isinstance(sam_config, dict):
            config_dict = sam_config
        elif isinstance(sam_config, str):
            config_dict = {sam_config: sam_config}
        else:
            raise ValueError(
                "Cannot parse SAM configs from {}".format(type(sam_config))
            )

        return SAMConfig(config_dict)

    @staticmethod
    def _parse_curtailment(curtailment_input):
        """Parse curtailment config object.

        Parameters
        ----------
        curtailment_input : None | dict | str | config.curtailment.Curtailment
            Inputs for curtailment parameters. If not None, curtailment inputs
            are expected. Can be:
                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config json file with path (str)
                - Instance of curtailment config object
                  (config.curtailment.Curtailment)

        Returns
        -------
        curtailments : NoneType | reV.config.curtailment.Curtailment
            None if no curtailment, reV curtailment config object if
            curtailment is being assessed.
        """
        if isinstance(curtailment_input, (str, dict)):
            # pointer to config file or explicit input namespace,
            # instantiate curtailment config object
            curtailment = Curtailment(curtailment_input)

        elif isinstance(curtailment_input, (Curtailment, type(None))):
            # pre-initialized curtailment object or no curtailment (None)
            curtailment = curtailment_input

        else:
            curtailment = None
            warn(
                "Curtailment inputs not recognized. Received curtailment "
                'input of type: "{}". Expected None, dict, str, or '
                "Curtailment object. Defaulting to no curtailment.",
                ConfigWarning,
            )

        return curtailment

    def index(self, gid):
        """Get the index location (iloc not loc) for a resource gid found in
        the project points.

        Parameters
        ----------
        gid : int
            Resource GID found in the project points gid column.

        Returns
        -------
        ind : int
            Row index of gid in the project points dataframe.
        """
        if gid not in self._df[SiteDataField.GID].values:
            e = (
                "Requested resource gid {} is not present in the project "
                "points dataframe. Cannot return row index.".format(gid)
            )
            logger.error(e)
            raise ConfigError(e)

        ind = np.where(self._df[SiteDataField.GID] == gid)[0][0]

        return ind

    def _check_points_config_mapping(self):
        """
        Check to ensure the project points (df) and SAM configs
        (sam_config_obj) are compatible. Update as necessary or break
        """
        # Extract unique config refences from project_points DataFrame
        df_configs = self.df[SiteDataField.CONFIG].unique()
        sam_configs = self.sam_inputs

        # Checks to make sure that the same number of SAM config files
        # as references in project_points DataFrame
        if len(df_configs) > len(sam_configs):
            msg = (
                "Points references {} configs while only "
                "{} SAM configs were provided!".format(
                    len(df_configs), len(sam_configs)
                )
            )
            logger.error(msg)
            raise ConfigError(msg)

        if len(df_configs) == 1 and df_configs[0] is None:
            self._df[SiteDataField.CONFIG] = list(sam_configs)[0]
            df_configs = self.df[SiteDataField.CONFIG].unique()

        # Check to see if config references in project_points DataFrame
        # are valid file paths, if compare with SAM configs
        # and update as needed
        configs = {}
        for config in df_configs:
            if os.path.isfile(config):
                configs[config] = config
            elif config in sam_configs:
                configs[config] = sam_configs[config]
            else:
                msg = "{} does not map to a valid configuration file".format(
                    config
                )
                logger.error(msg)
                raise ConfigError(msg)

        # If configs has any keys that are not in sam_configs then
        # something really weird happened so raise an error.
        if any(set(configs) - set(sam_configs)):
            msg = (
                "A wild config has appeared! Requested config keys for "
                "ProjectPoints are {} and previous config keys are {}".format(
                    list(configs), list(sam_configs)
                )
            )
            logger.error(msg)
            raise ConfigError(msg)

    def join_df(self, df2, key=SiteDataField.GID):
        """Join new df2 to the _df attribute using the _df's gid as pkey.

        This can be used to add site-specific data to the project_points,
        taking advantage of the points_control iterator/split functions such
        that only the relevant site data is passed to the analysis functions.

        Parameters
        ----------
        df2 : pd.DataFrame
            Dataframe to be joined to the self._df attribute (this instance
            of project points dataframe). This likely contains
            site-specific inputs that are to be passed to parallel workers.
        key : str
            Primary key of df2 to be joined to the _df attribute (this
            instance of the project points dataframe). Primary key
            of the self._df attribute is fixed as the gid column.
        """
        # ensure df2 doesnt have any duplicate columns for suffix reasons.
        df2_cols = [c for c in df2.columns if c not in self._df or c == key]
        self._df = pd.merge(
            self._df,
            df2[df2_cols],
            how="left",
            left_on=SiteDataField.GID,
            right_on=key,
            copy=False,
            validate="1:1",
        )

    def get_sites_from_config(self, config):
        """Get a site list that corresponds to a config key.

        Parameters
        ----------
        config : str
            SAM configuration ID associated with sites.

        Returns
        -------
        sites : list
            List of sites associated with the requested configuration ID. If
            the configuration ID is not recognized, an empty list is returned.
        """
        sites = self.df.loc[
            (self.df[SiteDataField.CONFIG] == config), SiteDataField.GID
        ].values

        return list(sites)

    @classmethod
    def split(cls, i0, i1, project_points):
        """Return split instance of a ProjectPoints instance w/ site subset.

        Parameters
        ----------
        i0 : int
            Starting INDEX (not resource gid) (inclusive) of the site property
            attribute to include in the split instance. This is not necessarily
            the same as the starting site number, for instance if ProjectPoints
            is sites 20:100, i0=0 i1=10 will result in sites 20:30.
        i1 : int
            Ending INDEX (not resource gid) (exclusive) of the site property
            attribute to include in the split instance. This is not necessarily
            the same as the final site number, for instance if ProjectPoints is
            sites 20:100, i0=0 i1=10 will result in sites 20:30.
        project_points: ProjectPoints
            Instance of project points to split.

        Returns
        -------
        sub : ProjectPoints
            New instance of ProjectPoints with a subset of the following
            attributes: sites, project points df, and the self dictionary data
            struct.
        """
        # Extract DF subset with only index values between i0 and i1
        n = len(project_points)
        if i0 > n or i1 > n:
            raise ValueError(
                "{} and {} must be within the range of "
                "project_points (0 - {})".format(i0, i1, n - 1)
            )

        points_df = project_points.df.iloc[i0:i1]

        # make a new instance of ProjectPoints with subset DF
        sub = cls(
            points_df,
            project_points.sam_config_obj,
            project_points.tech,
            curtailment=project_points.curtailment,
        )

        return sub

    @staticmethod
    def _parse_lat_lons(lat_lons):
        msg = (
            "Expecting a pair or multiple pairs of latitude and "
            "longitude coordinates!"
        )
        if isinstance(lat_lons, str):
            lat_lons = parse_table(lat_lons)
            cols = [
                c for c in lat_lons if c.lower().startswith(("lat", "lon"))
            ]
            lat_lons = lat_lons[sorted(cols)].values
        elif isinstance(lat_lons, (list, tuple)):
            lat_lons = np.array(lat_lons)
        elif isinstance(lat_lons, (int, float)):
            msg += " Recieved a single coordinate value!"
            logger.error(msg)
            raise ValueError(msg)

        if len(lat_lons.shape) == 1:
            lat_lons = np.expand_dims(lat_lons, axis=0)

        if lat_lons.shape[1] != 2:
            msg += " Received {} coordinate values!".format(lat_lons.shape[1])
            logger.error(msg)
            raise ValueError(msg)

        return lat_lons

    @classmethod
    def lat_lon_coords(
        cls, lat_lons, res_file, sam_configs, tech=None, curtailment=None
    ):
        """
        Generate ProjectPoints for gids nearest to given latitude longitudes

        Parameters
        ----------
        lat_lons : str | tuple | list | ndarray
            Pair or pairs of latitude longitude coordinates
        res_file : str
            Resource file, needed to fine nearest neighbors
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.
        tech : str, optional
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed,
            by default None
        curtailment : NoneType | dict | str | config.curtailment.Curtailment
            Inputs for curtailment parameters. If not None, curtailment inputs
            are expected. Can be:

                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config json file with path (str)
                - Instance of curtailment config object
                  (config.curtailment.Curtailment)


        Returns
        -------
        pp : ProjectPoints
            Initialized ProjectPoints object for points nearest to given
            lat_lons
        """
        lat_lons = cls._parse_lat_lons(lat_lons)

        multi_h5_res, hsds = check_res_file(res_file)
        if multi_h5_res:
            res_cls = MultiFileResourceX
            res_kwargs = {}
        else:
            res_cls = ResourceX
            res_kwargs = {"hsds": hsds}

        logger.info(
            "Converting latitude longitude coordinates into nearest "
            "ProjectPoints"
        )
        logger.debug("- (lat, lon) pairs:\n{}".format(lat_lons))
        with res_cls(res_file, **res_kwargs) as f:
            gids = f.lat_lon_gid(lat_lons)  # pylint: disable=no-member

        if isinstance(gids, int):
            gids = [gids]
        else:
            if len(gids) != len(np.unique(gids)):
                uniques, pos, counts = np.unique(
                    gids, return_counts=True, return_inverse=True
                )
                duplicates = {}
                for idx in np.where(counts > 1)[0]:
                    duplicate_lat_lons = lat_lons[np.where(pos == idx)[0]]
                    duplicates[uniques[idx]] = duplicate_lat_lons

                msg = (
                    "reV Cannot currently handle duplicate Resource gids! "
                    "The given latitude and longitudes map to the same "
                    "gids:\n{}".format(duplicates)
                )
                logger.error(msg)
                raise RuntimeError(msg)

            gids = gids.tolist()

        logger.debug("- Resource gids:\n{}".format(gids))

        pp = cls(
            gids,
            sam_configs,
            tech=tech,
            res_file=res_file,
            curtailment=curtailment,
        )

        if "points_order" in pp.df:
            lat_lons = lat_lons[pp.df["points_order"].values]

        pp._df["latitude"] = lat_lons[:, 0]
        pp._df["longitude"] = lat_lons[:, 1]

        return pp

    @classmethod
    def regions(
        cls, regions, res_file, sam_configs, tech=None, curtailment=None
    ):
        """
        Generate ProjectPoints for gids nearest to given latitude longitudes

        Parameters
        ----------
        regions : dict
            Dictionary of regions to extract points for in the form:
            {'region': 'region_column'}
        res_file : str
            Resource file, needed to fine nearest neighbors
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.
        tech : str, optional
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed,
            by default None
        curtailment : NoneType | dict | str | config.curtailment.Curtailment
            Inputs for curtailment parameters. If not None, curtailment inputs
            are expected. Can be:

                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config json file with path (str)
                - Instance of curtailment config object
                  (config.curtailment.Curtailment)


        Returns
        -------
        pp : ProjectPoints
            Initialized ProjectPoints object for points nearest to given
            lat_lons
        """
        multi_h5_res, hsds = check_res_file(res_file)
        if multi_h5_res:
            res_cls = MultiFileResourceX
        else:
            res_cls = ResourceX

        logger.info("Extracting ProjectPoints for desired regions")
        points = []
        with res_cls(res_file, hsds=hsds) as f:
            meta = f.meta
            for region, region_col in regions.items():
                logger.debug("- {}: {}".format(region_col, region))
                # pylint: disable=no-member
                gids = f.region_gids(region, region_col=region_col)
                logger.debug("- Resource gids:\n{}".format(gids))
                if points:
                    duplicates = np.intersect1d(gids, points).tolist()
                    if duplicates:
                        msg = (
                            "reV Cannot currently handle duplicate "
                            "Resource gids! The given regions containg the "
                            "same gids:\n{}".format(duplicates)
                        )
                        logger.error(msg)
                        raise RuntimeError(msg)

                points.extend(gids.tolist())

        pp = cls(
            points,
            sam_configs,
            tech=tech,
            res_file=res_file,
            curtailment=curtailment,
        )

        meta = meta.loc[pp.sites]
        cols = list(set(regions.values()))
        for c in cols:
            pp._df[c] = meta[c].values

        return pp



# -*- coding: utf-8 -*-
"""reV aggregation framework."""

import logging
import os
from abc import ABC, abstractmethod

import h5py
import numpy as np
import pandas as pd
from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

from reV.handlers.exclusions import ExclusionLayers
from reV.handlers.outputs import Outputs
from reV.supply_curve.exclusions import ExclusionMaskFromDict
from reV.supply_curve.extent import SupplyCurveExtent
from reV.supply_curve.points import AggregationSupplyCurvePoint
from reV.supply_curve.tech_mapping import TechMapping
from reV.utilities import ResourceMetaField, SupplyCurveField, log_versions
from reV.utilities.exceptions import (
    EmptySupplyCurvePointError,
    FileInputError,
    SupplyCurveInputError,
)

logger = logging.getLogger(__name__)


class AbstractAggFileHandler(ABC):
    """Simple framework to handle aggregation file context managers."""

    def __init__(
        self,
        excl_fpath,
        excl_dict=None,
        area_filter_kernel="queen",
        min_area=None,
    ):
        """
        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default 'queen'
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None
        """
        self._excl_fpath = excl_fpath
        self._excl = ExclusionMaskFromDict(
            excl_fpath,
            layers_dict=excl_dict,
            min_area=min_area,
            kernel=area_filter_kernel,
        )

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None:
            raise

    @abstractmethod
    def close(self):
        """Close all file handlers."""
        self._excl.close()

    @property
    def exclusions(self):
        """Get the exclusions file handler object.

        Returns
        -------
        _excl : ExclusionMask
            Exclusions h5 handler object.
        """
        return self._excl

    @property
    def h5(self):
        """
        Placeholder for h5 Resource handler
        """


class AggFileHandler(AbstractAggFileHandler):
    """
    Framework to handle aggregation file context manager:
    - exclusions .h5 file
    - h5 file to be aggregated
    """

    DEFAULT_H5_HANDLER = Resource

    def __init__(
        self,
        excl_fpath,
        h5_fpath,
        excl_dict=None,
        area_filter_kernel="queen",
        min_area=None,
        h5_handler=None,
    ):
        """
        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        h5_fpath : str
            Filepath to .h5 file to be aggregated
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default 'queen'
        min_area : float, optional
            Minimum required contiguous area filter in sq-km, by default None
        h5_handler : rex.Resource | None
            Optional special handler similar to the rex.Resource handler which
            is default.
        """
        super().__init__(
            excl_fpath,
            excl_dict=excl_dict,
            area_filter_kernel=area_filter_kernel,
            min_area=min_area,
        )

        if h5_handler is None:
            self._h5 = Resource(h5_fpath)
        else:
            self._h5 = h5_handler(h5_fpath)

    @property
    def h5(self):
        """
        Get the h5 file handler object.

        Returns
        -------
        _h5 : Outputs
            reV h5 outputs handler object.
        """
        return self._h5

    def close(self):
        """Close all file handlers."""
        self._excl.close()
        self._h5.close()


class BaseAggregation(ABC):
    """Abstract supply curve points aggregation framework based on only an
    exclusion file and techmap."""

    def __init__(
        self,
        excl_fpath,
        tm_dset,
        excl_dict=None,
        area_filter_kernel="queen",
        min_area=None,
        resolution=64,
        excl_area=None,
        res_fpath=None,
        gids=None,
        pre_extract_inclusions=False,
    ):
        """
        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        tm_dset : str
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data.
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default "queen"
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default None
        excl_area : float, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath, by default None
        gids : list, optional
            List of supply curve point gids to get summary for (can use to
            subset if running in parallel), or None for all gids in the SC
            extent, by default None
        pre_extract_inclusions : bool, optional
            Optional flag to pre-extract/compute the inclusion mask from the
            provided excl_dict, by default False. Typically faster to compute
            the inclusion mask on the fly with parallel workers.
        """
        self._excl_fpath = excl_fpath
        self._tm_dset = tm_dset
        self._excl_dict = excl_dict
        self._resolution = resolution
        self._area_filter_kernel = area_filter_kernel
        self._min_area = min_area
        self._res_fpath = res_fpath
        self._gids = gids
        self._pre_extract_inclusions = pre_extract_inclusions
        self._excl_area = self._get_excl_area(excl_fpath, excl_area=excl_area)
        self._shape = None

        self._validate_tech_mapping()

        if pre_extract_inclusions:
            self._inclusion_mask = (
                ExclusionMaskFromDict.extract_inclusion_mask(
                    excl_fpath,
                    tm_dset,
                    excl_dict=excl_dict,
                    area_filter_kernel=area_filter_kernel,
                    min_area=min_area,
                )
            )
        else:
            self._inclusion_mask = None

    def _validate_tech_mapping(self):
        """Check that tech mapping exists and create it if it doesn't"""

        with ExclusionLayers(self._excl_fpath) as f:
            dsets = f.h5.dsets

        excl_fp_is_str = isinstance(self._excl_fpath, str)
        tm_in_excl = self._tm_dset in dsets
        if tm_in_excl:
            logger.info('Found techmap "{}".'.format(self._tm_dset))
        elif not tm_in_excl and not excl_fp_is_str:
            msg = (
                'Could not find techmap dataset "{}" and cannot run '
                "techmap with arbitrary multiple exclusion filepaths "
                "to write to: {}".format(self._tm_dset, self._excl_fpath)
            )
            logger.error(msg)
            raise RuntimeError(msg)
        else:
            logger.info(
                'Could not find techmap "{}". Running techmap module.'.format(
                    self._tm_dset
                )
            )
            try:
                TechMapping.run(
                    self._excl_fpath, self._res_fpath, dset=self._tm_dset
                )
            except Exception as e:
                msg = (
                    "TechMapping process failed. Received the "
                    "following error:\n{}".format(e)
                )
                logger.exception(msg)
                raise RuntimeError(msg) from e

    @property
    def gids(self):
        """
        1D array of supply curve point gids to aggregate

        Returns
        -------
        ndarray
        """
        if self._gids is None:
            with SupplyCurveExtent(
                self._excl_fpath, resolution=self._resolution
            ) as sc:
                self._gids = sc.valid_sc_points(self._tm_dset)
        elif np.issubdtype(type(self._gids), np.number):
            self._gids = np.array([self._gids])
        elif not isinstance(self._gids, np.ndarray):
            self._gids = np.array(self._gids)

        return self._gids

    @property
    def shape(self):
        """Get the shape of the full exclusions raster.

        Returns
        -------
        tuple
        """
        if self._shape is None:
            with SupplyCurveExtent(
                self._excl_fpath, resolution=self._resolution
            ) as sc:
                self._shape = sc.exclusions.shape

        return self._shape

    @staticmethod
    def _get_excl_area(excl_fpath, excl_area=None):
        """
        Get exclusion area from excl_fpath pixel area. Confirm that the
        exclusion area is not None.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        excl_area : float, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath, by default None

        Returns
        -------
        excl_area : float
            Area of an exclusion pixel in km2
        """
        if excl_area is None:
            logger.debug(
                "Setting the exclusion area from the area of a pixel "
                "in {}".format(excl_fpath)
            )
            with ExclusionLayers(excl_fpath) as excl:
                excl_area = excl.pixel_area

        if excl_area is None:
            e = (
                "No exclusion pixel area was input and could not parse "
                "area from the exclusion file attributes!"
            )
            logger.error(e)
            raise SupplyCurveInputError(e)

        return excl_area

    @staticmethod
    def _check_inclusion_mask(inclusion_mask, gids, excl_shape):
        """
        Check inclusion mask to ensure it has the proper shape

        Parameters
        ----------
        inclusion_mask : np.ndarray | dict | optional
            2D array pre-extracted inclusion mask where 1 is included and 0 is
            excluded. This must be either match the full exclusion shape or
            be a dict lookup of single-sc-point exclusion masks corresponding
            to the gids input and keyed by gids, by default None which will
            calculate exclusions on the fly for each sc point.
        gids : list | ndarray
            sc point gids corresponding to inclusion mask
        excl_shape : tuple
            Full exclusion layers shape
        """
        if isinstance(inclusion_mask, dict):
            assert len(inclusion_mask) == len(gids)
        elif isinstance(inclusion_mask, np.ndarray):
            assert inclusion_mask.shape == excl_shape
        elif inclusion_mask is not None:
            msg = (
                "Expected inclusion_mask to be dict or array but received "
                "{}".format(type(inclusion_mask))
            )
            logger.error(msg)
            raise SupplyCurveInputError(msg)

    @staticmethod
    def _get_gid_inclusion_mask(
        inclusion_mask, gid, slice_lookup, resolution=64
    ):
        """
        Get inclusion mask for desired gid

        Parameters
        ----------
        inclusion_mask : np.ndarray | dict | optional
            2D array pre-extracted inclusion mask where 1 is included and 0 is
            excluded. This must be either match the full exclusion shape or
            be a dict lookup of single-sc-point exclusion masks corresponding
            to the gids input and keyed by gids, by default None which will
            calculate exclusions on the fly for each sc point.
        gid : int
            sc_point_gid value, used to extract inclusion mask from 2D
            inclusion array
        slice_lookup : dict
            Mapping of sc_point_gids to exclusion/inclusion row and column
            slices
        resolution : int, optional
            supply curve extent resolution, by default 64

        Returns
        -------
        gid_inclusions : ndarray | None
            2D array of inclusions for desired gid, normalized from 0, excluded
            to 1 fully included, if inclusion mask is None gid_inclusions
            is None
        """
        gid_inclusions = None
        if isinstance(inclusion_mask, dict):
            gid_inclusions = inclusion_mask[gid]
            assert gid_inclusions.shape[0] <= resolution
            assert gid_inclusions.shape[1] <= resolution
        elif isinstance(inclusion_mask, np.ndarray):
            row_slice, col_slice = slice_lookup[gid]
            gid_inclusions = inclusion_mask[row_slice, col_slice]
        elif inclusion_mask is not None:
            msg = (
                "Expected inclusion_mask to be dict or array but received "
                "{}".format(type(inclusion_mask))
            )
            logger.error(msg)
            raise SupplyCurveInputError(msg)

        return gid_inclusions

    @staticmethod
    def _parse_gen_index(gen_fpath):
        """Parse gen outputs for an array of generation gids corresponding to
        the resource gids.

        Parameters
        ----------
        gen_fpath : str
            Filepath to reV generation output .h5 file. This can also be a csv
            filepath to a project points input file.

        Returns
        -------
        gen_index : np.ndarray
            Array of generation gids with array index equal to resource gid.
            Array value is -1 if the resource index was not used in the
            generation run.
        """

        if gen_fpath.endswith(".h5"):
            with Resource(gen_fpath) as f:
                gen_index = f.meta
        elif gen_fpath.endswith(".csv"):
            gen_index = pd.read_csv(gen_fpath)
        else:
            msg = (
                "Could not recognize gen_fpath input, needs to be reV gen "
                "output h5 or project points csv but received: {}".format(
                    gen_fpath
                )
            )
            logger.error(msg)
            raise FileInputError(msg)

        if ResourceMetaField.GID in gen_index:
            gen_index = gen_index.rename(
                columns={ResourceMetaField.GID: SupplyCurveField.RES_GIDS}
            )
            gen_index[SupplyCurveField.GEN_GIDS] = gen_index.index
            gen_index = gen_index[
                [SupplyCurveField.RES_GIDS, SupplyCurveField.GEN_GIDS]
            ]
            gen_index = gen_index.set_index(keys=SupplyCurveField.RES_GIDS)
            gen_index = gen_index.reindex(
                range(int(gen_index.index.max() + 1))
            )
            gen_index = gen_index[SupplyCurveField.GEN_GIDS].values
            gen_index[np.isnan(gen_index)] = -1
            gen_index = gen_index.astype(np.int32)
        else:
            gen_index = None

        return gen_index


class Aggregation(BaseAggregation):
    """Concrete but generalized aggregation framework to aggregate ANY reV h5
    file to a supply curve grid (based on an aggregated exclusion grid)."""

    def __init__(
        self,
        excl_fpath,
        tm_dset,
        *agg_dset,
        excl_dict=None,
        area_filter_kernel="queen",
        min_area=None,
        resolution=64,
        excl_area=None,
        gids=None,
        pre_extract_inclusions=False,
    ):
        """
        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        tm_dset : str
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data.
        agg_dset : str
            Dataset to aggreate, can supply multiple datasets. The datasets
            should be scalar values for each site. This method cannot aggregate
            timeseries data.
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default "queen"
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default None
        excl_area : float, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        gids : list, optional
            List of supply curve point gids to get summary for (can use to
            subset if running in parallel), or None for all gids in the SC
            extent, by default None
        pre_extract_inclusions : bool, optional
            Optional flag to pre-extract/compute the inclusion mask from the
            provided excl_dict, by default False. Typically faster to compute
            the inclusion mask on the fly with parallel workers.
        """
        log_versions(logger)
        logger.info("Initializing Aggregation...")
        logger.debug("Exclusion filepath: {}".format(excl_fpath))
        logger.debug("Exclusion dict: {}".format(excl_dict))

        super().__init__(
            excl_fpath,
            tm_dset,
            excl_dict=excl_dict,
            area_filter_kernel=area_filter_kernel,
            min_area=min_area,
            resolution=resolution,
            excl_area=excl_area,
            gids=gids,
            pre_extract_inclusions=pre_extract_inclusions,
        )

        if isinstance(agg_dset, str):
            agg_dset = (agg_dset,)

        self._agg_dsets = agg_dset

    def _check_files(self, h5_fpath):
        """Do a preflight check on input files"""

        if not os.path.exists(self._excl_fpath):
            raise FileNotFoundError(
                "Could not find required exclusions file: " "{}".format(
                    self._excl_fpath
                )
            )

        if not os.path.exists(h5_fpath):
            raise FileNotFoundError(
                "Could not find required h5 file: " "{}".format(h5_fpath)
            )

        with h5py.File(self._excl_fpath, "r") as f:
            if self._tm_dset not in f:
                raise FileInputError(
                    'Could not find techmap dataset "{}" '
                    "in exclusions file: {}".format(
                        self._tm_dset, self._excl_fpath
                    )
                )

        with Resource(h5_fpath) as f:
            for dset in self._agg_dsets:
                if dset not in f:
                    raise FileInputError(
                        'Could not find provided dataset "{}"'
                        " in h5 file: {}".format(dset, h5_fpath)
                    )

    @classmethod
    def run_serial(
        cls,
        excl_fpath,
        h5_fpath,
        tm_dset,
        *agg_dset,
        agg_method="mean",
        excl_dict=None,
        inclusion_mask=None,
        area_filter_kernel="queen",
        min_area=None,
        resolution=64,
        excl_area=0.0081,
        gids=None,
        gen_index=None,
    ):
        """
        Standalone method to aggregate - can be parallelized.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        h5_fpath : str
            Filepath to .h5 file to aggregate
        tm_dset : str
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data.
        agg_dset : str
            Dataset to aggreate, can supply multiple datasets. The datasets
            should be scalar values for each site. This method cannot aggregate
            timeseries data.
        agg_method : str, optional
            Aggregation method, either mean or sum/aggregate, by default "mean"
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            by default None
        inclusion_mask : np.ndarray, optional
            2D array pre-extracted inclusion mask where 1 is included and 0 is
            excluded. This must be either match the full exclusion shape or
            be a list of single-sc-point exclusion masks corresponding to the
            gids input, by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default "queen"
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default 0.0081
        excl_area : float, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        gids : list, optional
            List of supply curve point gids to get summary for (can use to
            subset if running in parallel), or None for all gids in the SC
            extent, by default None
        gen_index : np.ndarray, optional
            Array of generation gids with array index equal to resource gid.
            Array value is -1 if the resource index was not used in the
            generation run, by default None

        Returns
        -------
        agg_out : dict
            Aggregated values for each aggregation dataset
        """
        with SupplyCurveExtent(excl_fpath, resolution=resolution) as sc:
            exclusion_shape = sc.exclusions.shape
            if gids is None:
                gids = sc.valid_sc_points(tm_dset)
            elif np.issubdtype(type(gids), np.number):
                gids = [gids]

            slice_lookup = sc.get_slice_lookup(gids)

        cls._check_inclusion_mask(inclusion_mask, gids, exclusion_shape)

        # pre-extract handlers so they are not repeatedly initialized
        file_kwargs = {
            "excl_dict": excl_dict,
            "area_filter_kernel": area_filter_kernel,
            "min_area": min_area,
        }
        dsets = (
            *agg_dset,
            "meta",
        )
        agg_out = {ds: [] for ds in dsets}
        with AggFileHandler(excl_fpath, h5_fpath, **file_kwargs) as fh:
            n_finished = 0
            for gid in gids:
                gid_inclusions = cls._get_gid_inclusion_mask(
                    inclusion_mask, gid, slice_lookup, resolution=resolution
                )
                try:
                    gid_out = AggregationSupplyCurvePoint.run(
                        gid,
                        fh.exclusions,
                        fh.h5,
                        tm_dset,
                        *agg_dset,
                        agg_method=agg_method,
                        excl_dict=excl_dict,
                        inclusion_mask=gid_inclusions,
                        resolution=resolution,
                        excl_area=excl_area,
                        exclusion_shape=exclusion_shape,
                        close=False,
                        gen_index=gen_index,
                    )

                except EmptySupplyCurvePointError:
                    logger.debug(
                        "SC gid {} is fully excluded or does not "
                        "have any valid source data!".format(gid)
                    )
                except Exception as e:
                    msg = "SC gid {} failed!".format(gid)
                    logger.exception(msg)
                    raise RuntimeError(msg) from e
                else:
                    n_finished += 1
                    logger.debug(
                        "Serial aggregation: "
                        "{} out of {} points complete".format(
                            n_finished, len(gids)
                        )
                    )
                    log_mem(logger)
                    for k, v in gid_out.items():
                        agg_out[k].append(v)

        return agg_out

    def run_parallel(
        self,
        h5_fpath,
        agg_method="mean",
        excl_area=None,
        max_workers=None,
        sites_per_worker=100,
    ):
        """
        Aggregate in parallel

        Parameters
        ----------
        h5_fpath : str
            Filepath to .h5 file to aggregate
        agg_method : str, optional
            Aggregation method, either mean or sum/aggregate, by default "mean"
        excl_area : float, optional
            Area of an exclusion cell (square km), by default None
        max_workers : int, optional
            Number of cores to run summary on. None is all available cpus,
            by default None
        sites_per_worker : int, optional
            Number of SC points to process on a single parallel worker,
            by default 100

        Returns
        -------
        agg_out : dict
            Aggregated values for each aggregation dataset
        """

        self._check_files(h5_fpath)
        gen_index = self._parse_gen_index(h5_fpath)

        slice_lookup = None
        chunks = int(np.ceil(len(self.gids) / sites_per_worker))
        chunks = np.array_split(self.gids, chunks)

        if self._inclusion_mask is not None:
            with SupplyCurveExtent(
                self._excl_fpath, resolution=self._resolution
            ) as sc:
                assert sc.exclusions.shape == self._inclusion_mask.shape
                slice_lookup = sc.get_slice_lookup(self.gids)

        logger.info(
            "Running supply curve point aggregation for "
            "points {} through {} at a resolution of {} "
            "on {} cores in {} chunks.".format(
                self.gids[0],
                self.gids[-1],
                self._resolution,
                max_workers,
                len(chunks),
            )
        )

        n_finished = 0
        futures = []
        dsets = self._agg_dsets + ("meta",)
        agg_out = {ds: [] for ds in dsets}
        loggers = [__name__, "reV.supply_curve.points", "reV"]
        with SpawnProcessPool(max_workers=max_workers, loggers=loggers) as exe:
            # iterate through split executions, submitting each to worker
            for gid_set in chunks:
                # submit executions and append to futures list
                chunk_incl_masks = None
                if self._inclusion_mask is not None:
                    chunk_incl_masks = {}
                    for gid in gid_set:
                        rs, cs = slice_lookup[gid]
                        chunk_incl_masks[gid] = self._inclusion_mask[rs, cs]

                # submit executions and append to futures list
                futures.append(
                    exe.submit(
                        self.run_serial,
                        self._excl_fpath,
                        h5_fpath,
                        self._tm_dset,
                        *self._agg_dsets,
                        agg_method=agg_method,
                        excl_dict=self._excl_dict,
                        inclusion_mask=chunk_incl_masks,
                        area_filter_kernel=self._area_filter_kernel,
                        min_area=self._min_area,
                        resolution=self._resolution,
                        excl_area=excl_area,
                        gids=gid_set,
                        gen_index=gen_index,
                    )
                )

            # gather results
            for future in futures:
                n_finished += 1
                logger.info(
                    "Parallel aggregation futures collected: "
                    "{} out of {}".format(n_finished, len(chunks))
                )
                for k, v in future.result().items():
                    if v:
                        agg_out[k].extend(v)

        return agg_out

    def aggregate(
        self,
        h5_fpath,
        agg_method="mean",
        max_workers=None,
        sites_per_worker=100,
    ):
        """
        Aggregate with given agg_method

        Parameters
        ----------
        h5_fpath : str
            Filepath to .h5 file to aggregate
        agg_method : str, optional
            Aggregation method, either mean or sum/aggregate, by default "mean"
        max_workers : int, optional
            Number of cores to run summary on. None is all available cpus,
            by default None
        sites_per_worker : int, optional
            Number of SC points to process on a single parallel worker,
            by default 100

        Returns
        -------
        agg : dict
            Aggregated values for each aggregation dataset
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        if max_workers == 1:
            self._check_files(h5_fpath)
            gen_index = self._parse_gen_index(h5_fpath)
            agg = self.run_serial(
                self._excl_fpath,
                h5_fpath,
                self._tm_dset,
                *self._agg_dsets,
                agg_method=agg_method,
                excl_dict=self._excl_dict,
                gids=self.gids,
                inclusion_mask=self._inclusion_mask,
                area_filter_kernel=self._area_filter_kernel,
                min_area=self._min_area,
                resolution=self._resolution,
                excl_area=self._excl_area,
                gen_index=gen_index,
            )
        else:
            agg = self.run_parallel(
                h5_fpath=h5_fpath,
                agg_method=agg_method,
                excl_area=self._excl_area,
                max_workers=max_workers,
                sites_per_worker=sites_per_worker,
            )

        if not agg["meta"]:
            e = (
                "Supply curve aggregation found no non-excluded SC points. "
                "Please check your exclusions or subset SC GID selection."
            )
            logger.error(e)
            raise EmptySupplyCurvePointError(e)

        for k, v in agg.items():
            if k == "meta":
                v = pd.concat(v, axis=1).T
                v = v.sort_values(SupplyCurveField.SC_POINT_GID)
                v = v.reset_index(drop=True)
                v.index.name = SupplyCurveField.SC_GID
                agg[k] = v
            else:
                v = np.dstack(v)[0]
                if v.shape[0] == 1:
                    v = v.flatten()

                agg[k] = v

        return agg

    @staticmethod
    def save_agg_to_h5(h5_fpath, out_fpath, aggregation):
        """
        Save aggregated data to disc in .h5 format

        Parameters
        ----------
        out_fpath : str
            Output .h5 file path
        aggregation : dict
            Aggregated values for each aggregation dataset
        """
        agg_out = aggregation.copy()
        meta = agg_out.pop("meta").reset_index()
        for c in meta.columns:
            try:
                meta[c] = pd.to_numeric(meta[c])
            except (ValueError, TypeError):
                pass

        dsets = []
        shapes = {}
        attrs = {}
        chunks = {}
        dtypes = {}
        time_index = None
        with Resource(h5_fpath) as f:
            for dset, data in agg_out.items():
                dsets.append(dset)
                shape = data.shape
                shapes[dset] = shape
                if len(data.shape) == 2:
                    if ("time_index" in f) and (shape[0] == f.shape[0]):
                        if time_index is None:
                            time_index = f.time_index

                attrs[dset] = f.get_attrs(dset=dset)
                _, dtype, chunk = f.get_dset_properties(dset)
                chunks[dset] = chunk
                dtypes[dset] = dtype

        Outputs.init_h5(
            out_fpath,
            dsets,
            shapes,
            attrs,
            chunks,
            dtypes,
            meta,
            time_index=time_index,
        )

        with Outputs(out_fpath, mode="a") as out:
            for dset, data in agg_out.items():
                out[dset] = data

    @classmethod
    def run(
        cls,
        excl_fpath,
        h5_fpath,
        tm_dset,
        *agg_dset,
        excl_dict=None,
        area_filter_kernel="queen",
        min_area=None,
        resolution=64,
        excl_area=None,
        gids=None,
        pre_extract_inclusions=False,
        agg_method="mean",
        max_workers=None,
        sites_per_worker=100,
        out_fpath=None,
    ):
        """Get the supply curve points aggregation summary.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        h5_fpath : str
            Filepath to .h5 file to aggregate
        tm_dset : str
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data.
        agg_dset : str
            Dataset to aggreate, can supply multiple datasets. The datasets
            should be scalar values for each site. This method cannot aggregate
            timeseries data.
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default "queen"
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default None
        excl_area : float, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        gids : list, optional
            List of supply curve point gids to get summary for (can use to
            subset if running in parallel), or None for all gids in the SC
            extent, by default None
        pre_extract_inclusions : bool, optional
            Optional flag to pre-extract/compute the inclusion mask from the
            provided excl_dict, by default False. Typically faster to compute
            the inclusion mask on the fly with parallel workers.
        agg_method : str, optional
            Aggregation method, either mean or sum/aggregate, by default "mean"
        max_workers : int, optional
            Number of cores to run summary on. None is all available cpus,
            by default None
        sites_per_worker : int, optional
            Number of SC points to process on a single parallel worker,
            by default 100
        out_fpath : str, optional
            Output .h5 file path, by default None

        Returns
        -------
        agg : dict
            Aggregated values for each aggregation dataset
        """

        agg = cls(
            excl_fpath,
            tm_dset,
            *agg_dset,
            excl_dict=excl_dict,
            area_filter_kernel=area_filter_kernel,
            min_area=min_area,
            resolution=resolution,
            excl_area=excl_area,
            gids=gids,
            pre_extract_inclusions=pre_extract_inclusions,
        )

        aggregation = agg.aggregate(
            h5_fpath=h5_fpath,
            agg_method=agg_method,
            max_workers=max_workers,
            sites_per_worker=sites_per_worker,
        )

        if out_fpath is not None:
            agg.save_agg_to_h5(h5_fpath, out_fpath, aggregation)

        return aggregation



# -*- coding: utf-8 -*-
"""
Multi-year means CLI utility functions.
"""
import logging
from gaps.cli import as_click_command, CLICommandFromFunction

from reV.handlers.multi_year import my_collect_groups
from reV.utilities import ModuleName


logger = logging.getLogger(__name__)


my_command = CLICommandFromFunction(my_collect_groups,
                                    name=str(ModuleName.MULTI_YEAR),
                                    split_keys=None)
main = as_click_command(my_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Multi-Year collect CLI.')
        raise



# -*- coding: utf-8 -*-
"""Functions used for pytests"""

import os

import numpy as np
import pandas as pd
from packaging import version
from rex.outputs import Outputs as RexOutputs

from reV.utilities import ResourceMetaField


def pd_date_range(*args, **kwargs):
    """A simple wrapper on the pd.date_range() method that handles the closed
    vs. inclusive kwarg change in pd 1.4.0"""
    incl = version.parse(pd.__version__) >= version.parse('1.4.0')

    if incl and 'closed' in kwargs:
        kwargs['inclusive'] = kwargs.pop('closed')
    elif not incl and 'inclusive' in kwargs:
        kwargs['closed'] = kwargs.pop('inclusive')
        if kwargs['closed'] == 'both':
            kwargs['closed'] = None

    return pd.date_range(*args, **kwargs)


def write_chunk(meta, times, data, features, out_file):
    """Write data chunk to an h5 file

    Parameters
    ----------
    meta : dict
        Dictionary of meta data for this chunk. Includes flattened lat and lon
        arrays
    times : pd.DatetimeIndex
        times in this chunk
    features : list
        List of feature names in this chunk
    out_file : str
        Name of output file
    """
    with RexOutputs(out_file, 'w') as fh:
        fh.meta = meta
        fh.time_index = times
        for feature in features:
            flat_data = data.reshape((-1, len(times)))
            flat_data = np.transpose(flat_data, (1, 0))
            fh.add_dataset(out_file, feature, flat_data, dtype=np.float32)


def make_fake_h5_chunks(td, features, shuffle=False):
    """Make fake h5 chunks to test collection

    Parameters
    ----------
    td : tempfile.TemporaryDirectory
        Test TemporaryDirectory
    features : list
        List of dsets to write to chunks
    shuffle : bool
        Whether to shuffle gids

    Returns
    -------
    out_pattern : str
        Pattern for output file names
    data : ndarray
        Full non-chunked data array
    features : list
        List of feature names in output
    s_slices : list
        List of spatial slices used to chunk full data array
    times : pd.DatetimeIndex
        Times in output
    """
    shape = (50, 50, 48)
    data = np.random.uniform(0, 20, shape)
    lat = np.linspace(90, 0, 50)
    lon = np.linspace(-180, 0, 50)
    lon, lat = np.meshgrid(lon, lat)
    gids = np.arange(np.product(lat.shape))
    if shuffle:
        np.random.shuffle(gids)
    gids = gids.reshape(shape[:-1])
    times = pd_date_range('20220101', '20220103', freq='3600s',
                          inclusive='left')
    s_slices = [slice(0, 25), slice(25, 50)]
    out_pattern = os.path.join(td, 'chunks_{i}_{j}.h5')

    for i, s1 in enumerate(s_slices):
        for j, s2 in enumerate(s_slices):
            out_file = out_pattern.format(i=i, j=j)
            meta = pd.DataFrame(
                {ResourceMetaField.LATITUDE: lat[s1, s2].flatten(),
                 ResourceMetaField.LONGITUDE: lon[s1, s2].flatten(),
                 ResourceMetaField.GID: gids[s1, s2].flatten()})
            write_chunk(meta=meta, times=times, data=data[s1, s2],
                        features=features, out_file=out_file)

    out = (out_pattern.format(i='*', j='*'), data, features, s_slices, times)
    return out



# -*- coding: utf-8 -*-
"""
reV Supply Curve Aggregation CLI utility functions.
"""
import os
import logging

from rex.multi_file_resource import MultiFileResource
from gaps.cli import as_click_command, CLICommandFromClass

from reV.supply_curve.sc_aggregation import SupplyCurveAggregation
from reV.utilities import ModuleName
from reV.utilities.cli_functions import parse_from_pipeline
from reV.utilities.exceptions import ConfigError


logger = logging.getLogger(__name__)


def _preprocessor(config, out_dir):
    """Preprocess aggregation config user input.

    Parameters
    ----------
    config : dict
        User configuration file input as (nested) dict.
    out_dir : str
        Path to output file directory.

    Returns
    -------
    dict
        Updated config file.
    """
    config = _format_res_fpath(config)
    _validate_tm(config)

    key_to_modules = {"gen_fpath": [ModuleName.MULTI_YEAR,
                                    ModuleName.COLLECT,
                                    ModuleName.GENERATION],
                      "econ_fpath": [ModuleName.MULTI_YEAR,
                                     ModuleName.COLLECT,
                                     ModuleName.ECON]}
    for key, modules in key_to_modules.items():
        config = parse_from_pipeline(config, out_dir, key, modules)

    return config


def _format_res_fpath(config):
    """Format res_fpath with year, if need be. """
    res_fpath = config.setdefault("res_fpath", None)
    if isinstance(res_fpath, str):
        if '{}' in res_fpath:
            for year in range(1950, 2100):
                if os.path.exists(res_fpath.format(year)):
                    break
            else:
                msg = ("Could not find any files that match the pattern"
                       "{!r}".format(res_fpath.format("<year>")))
                logger.error(msg)
                raise FileNotFoundError(msg)

            res_fpath = res_fpath.format(year)

        elif not os.path.exists(res_fpath):
            msg = "Could not find resource file: {!r}".format(res_fpath)
            logger.error(msg)
            raise FileNotFoundError(msg)

        config["res_fpath"] = res_fpath

    return config


def _validate_tm(config):
    """Check that tm_dset exists or that res_fpath is given (to generate tm)"""
    paths = config["excl_fpath"]
    if isinstance(paths, str):
        paths = [paths]

    with MultiFileResource(paths, check_files=False) as res:
        dsets = res.datasets

    if config["tm_dset"] not in dsets and config["res_fpath"] is None:
        raise ConfigError('Techmap dataset "{}" not found in exclusions '
                          'file, resource file input "res_fpath" is '
                          'required to create the techmap file.'
                          .format(config["tm_dset"]))


NAME = str(ModuleName.SUPPLY_CURVE_AGGREGATION)
sc_agg_command = CLICommandFromClass(SupplyCurveAggregation, method="run",
                                     name=NAME, add_collect=False,
                                     split_keys=None,
                                     config_preprocessor=_preprocessor)
main = as_click_command(sc_agg_command)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV SC Aggregation CLI.')
        raise



# -*- coding: utf-8 -*-
"""
turbine packing module.
"""
import numpy as np
from shapely.geometry import Polygon, MultiPolygon, Point
from reV.bespoke.plotting_functions import get_xy
from reV.utilities.exceptions import WhileLoopPackingError


class PackTurbines():
    """Framework to maximize plant capacity in a provided wind plant area.
    """

    def __init__(self, min_spacing, safe_polygons, weight_x=0.0013547):
        """
        Parameters
        ----------
        min_spacing : float
            The minimum allowed spacing between wind turbines.
        safe_polygons : Polygon | MultiPolygon
            The "safe" area(s) where turbines can be placed without
            violating boundary, setback, exclusion, or other constraints.
        weight_x : float, optional
        """

        self.min_spacing = min_spacing
        self.safe_polygons = safe_polygons
        self.weight_x = weight_x

        # turbine locations
        self.turbine_x = np.array([])
        self.turbine_y = np.array([])

    def pack_turbines_poly(self):
        """Fast packing algorithm that maximizes plant capacity in a
        provided wind plant area. Sets the the optimal locations to
        self.turbine_x and self.turbine_y
        """

        if self.safe_polygons.area > 0.0:
            can_add_more = True
            leftover = MultiPolygon(self.safe_polygons)
            iters = 0
            while can_add_more:
                iters += 1
                if iters > 10000:
                    msg = ('Too many points placed in packing algorithm')
                    raise WhileLoopPackingError(msg)

                if leftover.area > 0:
                    nareas = len(leftover.geoms)
                    areas = np.zeros(len(leftover.geoms))
                    for i in range(nareas):
                        areas[i] = leftover.geoms[i].area
                    m = min(i for i in areas if i > 0)
                    ind = np.where(areas == m)[0][0]
                    # smallest_area = leftover.geoms[np.argmin(areas)]
                    smallest_area = leftover.geoms[ind]
                    exterior_coords = smallest_area.exterior.coords[:]
                    x, y = get_xy(exterior_coords)
                    metric = self.weight_x * x + y
                    index = np.argmin(metric)
                    self.turbine_x = np.append(self.turbine_x,
                                               x[index])
                    self.turbine_y = np.append(self.turbine_y,
                                               y[index])
                    new_turbine = Point(x[index],
                                        y[index]
                                        ).buffer(self.min_spacing)
                else:
                    break
                leftover = leftover.difference(new_turbine)
                if isinstance(leftover, Polygon):
                    leftover = MultiPolygon([leftover])

    def clear(self):
        """Reset the packing algorithm by clearing the x and y turbine arrays
        """
        self.turbine_x = np.array([])
        self.turbine_y = np.array([])


def smallest_area_with_tiebreakers(g):
    """_summary_

    This function helps break ties in the area of two different
    geometries using their exterior coordinate values.

    Parameters
    ----------
    g : _type_
        A geometry object with an `area` and an
        `exterior.coords` coords attribute.

    Returns
    -------
    tuple
        Tuple with the following elements:
            - area of the geometry
            - minimum exterior coordinate (southwest)
            - maximum exterior coordinate (northeast)
    """
    return g.area, min(g.exterior.coords), max(g.exterior.coords)
